import logging
import os
import re
import sys
import tempfile
from logging.handlers import RotatingFileHandler
from typing import Optional

from aiko_services.main.utilities.logger import (
    _LOG_FORMAT_DATETIME,
    _LOG_FORMAT_DEFAULT,
)

__all__ = ["configure_root_logger"]


DEBUG_LOG_FORMAT = (
    "%(asctime)s.%(msecs)03d %(levelname)-8s %(threadName)s %(name)s.%(funcName)s:%(lineno)-6d %(message)s"
)


def is_running_under_pytest():
    return "pytest" in sys.modules


def configure_root_logger(
    _log_path: Optional[str] = None,
    _log_level: Optional[str] = None,
    _log_rotation_max_kilobytes: Optional[int] = None,
    _log_rotation_backup_count: Optional[int] = None,
):
    """Configure the root logger with file and stream handlers."""
    if _log_path is None:
        temp_file = tempfile.NamedTemporaryFile(suffix=".log", delete=False)
        _log_path = temp_file.name
        temp_file.close()
    _log_level = os.getenv("HL_LOG_LEVEL", "WARNING")

    # Set defaults for log rotation parameters
    if _log_rotation_max_kilobytes is None:
        _log_rotation_max_kilobytes = 100 * 1024  # 100 MB in KB
    if _log_rotation_backup_count is None:
        _log_rotation_backup_count = 4

    root = logging.getLogger()
    if root.hasHandlers():
        for handler in root.handlers:
            handler.close()
        root.handlers.clear()

    log_format = DEBUG_LOG_FORMAT if _log_level == "DEBUG" else _LOG_FORMAT_DEFAULT
    formatter = logging.Formatter(
        log_format,
        datefmt=_LOG_FORMAT_DATETIME,
    )

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)

    handlers = [stream_handler]

    log_level_mapping = {
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
    }
    log_level = log_level_mapping.get(_log_level)
    if log_level is None:
        raise SystemExit(f"Invalid log_level '{_log_level}'")

    if not is_running_under_pytest():
        # Ensure log_path exists
        directory = os.path.dirname(_log_path)
        os.makedirs(directory, exist_ok=True)

        if not os.path.exists(_log_path):
            with open(_log_path, "w") as file:
                file.write("")  # Creates an empty file

        # Setup File Handler
        file_handler = RotatingFileHandler(
            _log_path,
            maxBytes=_log_rotation_max_kilobytes * 1024,  # Convert KB to bytes
            backupCount=_log_rotation_backup_count,
        )
        file_handler.setFormatter(formatter)

        handlers.append(file_handler)
    logging.basicConfig(handlers=handlers)
    # Set the log level for highlighter code
    logging.getLogger("highlighter").setLevel(log_level)
    logging.getLogger(__name__).info(f"log_path: {_log_path}")

    for module_path, module_log_level in get_log_level_env_vars().items():
        logging.getLogger(module_path).setLevel(module_log_level)
        logging.getLogger(__name__).info(f"Set log level for {module_path} to {module_log_level}")


def get_log_level_env_vars():
    """Lookup log-level directives from environment variables
    matching the pattern:
        LOG_LEVEL_module_DOT_path_DOT_segment=INFO
    into {"module.path.segment": "INFO"}
    """
    pattern = re.compile(r"^LOG_LEVEL_(.*)$")
    results = {}
    for key, value in os.environ.items():
        match = pattern.match(key)
        if match:
            # Replace all "_DOT_" with "." in the matched part
            dotted_path = match.group(1).replace("_DOT_", ".")
            results[dotted_path] = value
    return results


class ColourStr:
    HEADER = "\033[95m"

    BLUE = "\033[94m"
    CYAN = "\033[96m"
    GREEN = "\033[92m"
    ORANGE = "\033[93m"
    RED = "\033[91m"

    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"

    RESET = "\033[0m"

    @staticmethod
    def blue(s):
        return ColourStr.BLUE + s + ColourStr.RESET

    @staticmethod
    def cyan(s):
        return ColourStr.CYAN + s + ColourStr.RESET

    @staticmethod
    def green(s):
        return ColourStr.GREEN + s + ColourStr.RESET

    @staticmethod
    def red(s):
        return ColourStr.RED + s + ColourStr.RESET

    @staticmethod
    def bold(s):
        return ColourStr.BOLD + s + ColourStr.RESET

    @staticmethod
    def underline(s):
        return ColourStr.UNDERLINE + s + ColourStr.RESET



import logging
from enum import Enum
from functools import lru_cache
from pathlib import Path
from typing import List, Optional, Union
from uuid import UUID
from warnings import warn

import numpy as np
import pandas as pd
import yaml
from PIL.Image import Image
from shapely import geometry as geom

from highlighter.client.io import _pil_open_image_path
from highlighter.core import (
    OBJECT_CLASS_ATTRIBUTE_UUID,
    PIXEL_LOCATION_ATTRIBUTE_UUID,
    LabeledUUID,
)
from highlighter.datasets.cropping import CropArgs, crop_rect_from_poly
from highlighter.datasets.formats.coco.common import (
    segmentation_to_bbox,
    shapely_to_segmentation,
)
from highlighter.datasets.interfaces import IWriter

__all__ = ["YoloWriter"]

PathLike = Union[str, Path]
logger = logging.getLogger(__name__)


@lru_cache()
def _read_image_lru(image_path) -> Image:
    return _pil_open_image_path(image_path)


class YoloWriter(IWriter):
    format_name = "yolo"
    annotation_count = -1
    image_count = -1

    class TASK(str, Enum):
        DETECT = "detect"
        SEGMENT = "segment"
        CLASSIFY = "classify"

    def __init__(
        self,
        output_dir: PathLike,
        image_cache_dir: PathLike,
        category_attribute_id: Union[str, UUID, LabeledUUID] = OBJECT_CLASS_ATTRIBUTE_UUID,
        categories: Optional[List[Union[str, UUID]]] = None,
        task: Union[str, TASK] = TASK.DETECT,
        crop_args: Optional[CropArgs] = None,
    ):
        """Save a Highlighter Dataset object as a YoloV8 Detection dataset

        The Highlighter Dataset object must have at least the split names
        "train" and "test" and optionally "val". If "val" is not a split
        "test" will be duplicated in the data.yaml

        Args:
            output_dir: Root level directory for the output dataset
            category_attribute_id: Which attribute to use for the detection categories
            categories: Select a subset of the categories. If None will use all
            image_cache_dir: Directory of locally stored images. The yolo
            <train|val|test>/images directories will contain symlinks to these files.
        """
        self.output_dir = Path(output_dir)
        self.category_attribute_id = category_attribute_id
        self.categories = categories
        self.image_cache_dir = Path(image_cache_dir).absolute()
        self.task: self.TASK = self.TASK(task)
        self.crop_args = crop_args

    @staticmethod
    def generate_data_config_dict(
        root_dir: Path,
        train_image_dir: Path,
        val_image_dir: Path,
        test_image_dir: Optional[Path],
        categories: List[str],
    ) -> dict:
        config = {
            "path": str(root_dir),
            "train": str(train_image_dir.relative_to(root_dir)),
            "val": str(val_image_dir.relative_to(root_dir)),
            "names": {i: name for i, name in enumerate(categories)},
            "nc": len(categories),
        }
        if test_image_dir is not None:
            config["test"] = str(test_image_dir.relative_to(root_dir))

        return config

    def _to_classify_label(
        self,
        category: Union[str, UUID],
        pixel_location: geom.Polygon,
        source_image_path: Path,
        split_dir: Path,
    ):
        image_pil = _read_image_lru(source_image_path)
        crop = crop_rect_from_poly(image_pil, pixel_location, crop_args=self.crop_args)
        l, t, r, b = [int(p) for p in pixel_location.bounds]
        dest_image_path = (
            split_dir
            / str(category)
            / f"{source_image_path.stem}-{l}-{t}-{r}-{b}{source_image_path.suffix.lower()}"
        )
        dest_image_path.parent.mkdir(exist_ok=True, parents=True)
        crop.save(dest_image_path)

    def _to_detect_label(self, cat_idx, segmentation, scale_wh):
        im_w, im_h = scale_wh

        box_left, box_top, box_w, box_h = segmentation_to_bbox(segmentation)
        box_cen_x = (box_left + box_w / 2) / im_w
        box_cen_y = (box_top + box_h / 2) / im_h
        box_w /= im_w
        box_h /= im_h
        return f"{cat_idx} {box_cen_x} {box_cen_y} {box_w} {box_h}"

    def _to_segment_label(self, cat_idx, segmentation, scale_wh):
        if len(segmentation) > 1:
            warn("YoloWriter only supports " "single polygon, not multipolygon.")
        segmentation_arr = np.array(segmentation[0])
        scale = 1 / np.array(scale_wh * (segmentation_arr.shape[0] // 2))
        scaled_segmentation = (segmentation_arr * scale).tolist()
        segmentation_str = " ".join([str(s) for s in scaled_segmentation])
        return f"{cat_idx} {segmentation_str}"

    def _assert_is_image_dataset(self, dataset):
        images_exts = (".jpg", ".jpeg", ".png")
        if any([Path(f).suffix.lower() not in images_exts for f in dataset.data_files_df.filename.unique()]):
            raise ValueError(
                "Dataset contains non image filenames. If you have a video dataset with keyframes use Dataset.interpolate_from_keyframes to convert to an image dataset"
            )

    def write(
        self,
        dataset: "Dataset",
    ):

        self._assert_is_image_dataset(dataset)

        unique_splits = dataset.data_files_df.split.unique()
        if not all([s in unique_splits for s in ("train", "val")]):
            raise ValueError(
                "data_files_df must have at split names "
                "['train', 'val'] optionally 'test'. Got: "
                f"{unique_splits}"
            )

        unique_categories = dataset.annotations_df[
            dataset.annotations_df.attribute_id == self.category_attribute_id
        ].value.unique()
        unique_categories.sort()
        unique_categories = unique_categories.tolist()

        # Allign categories with passed in categories
        # else fall back on sorted unique_categories
        if self.categories is not None:
            if not all([c in unique_categories for c in self.categories]):
                raise ValueError(
                    "All categories must appear in the source dataset "
                    f"got: categories = {self.categories} and "
                    f"source dataset categories = {unique_categories}"
                )
            categories_of_interest: List[str] = [str(c) for c in self.categories]
        else:
            categories_of_interest: List[str] = [str(c) for c in unique_categories]

        def to_yolo_annotation(grp, categories_of_interest=categories_of_interest):
            try:
                data = grp.set_index("entity_id").to_dict("index")
                data_file_id = grp.name

                labels = []
                entities = grp.set_index("entity_id").to_dict("index")
                for entity_id, data in entities.items():
                    pixel_loc_uuid_str = str(PIXEL_LOCATION_ATTRIBUTE_UUID)
                    if pixel_loc_uuid_str not in data:
                        raise ValueError(
                            f"each entity must have a {PIXEL_LOCATION_ATTRIBUTE_UUID} attribute, got: {data} for entity_id {entity_id}"
                        )

                    category_uuid_str = str(self.category_attribute_id)
                    if category_uuid_str not in data:
                        raise ValueError(
                            f"each entity must have a {category_uuid_str} attribute, got: {data} for entity_id {entity_id}"
                        )

                    category = data[category_uuid_str]
                    cat_idx = categories_of_interest.index(str(category))

                    im_w = data["width"]
                    im_h = data["height"]
                    pixel_location = data[pixel_loc_uuid_str]
                    segmentation = shapely_to_segmentation(pixel_location)
                    split_name = grp.split.values[0]
                    source_filename = Path(grp.filename.values[0])
                    source_filename = f"{source_filename.stem}{source_filename.suffix.lower()}"

                    if self.task in (self.TASK.DETECT, self.TASK.SEGMENT):
                        labels_dir = self.output_dir / "labels"
                        labels_dir.mkdir(parents=True, exist_ok=True)
                        to_label_fn = (
                            self._to_detect_label if self.task == self.TASK.DETECT else self._to_segment_label
                        )
                        labels.append(to_label_fn(cat_idx, segmentation, (im_w, im_h)))
                        labels_path = labels_dir / split_name / f"{data_file_id}.txt"
                        labels_path.parent.mkdir(exist_ok=True)
                        with labels_path.open("w") as f:
                            f.write("\n".join(labels))

                        image_symlink: Path = self.output_dir / "images" / split_name / source_filename
                        image_symlink.parent.mkdir(parents=True, exist_ok=True)
                        if not image_symlink.exists():
                            image_symlink.symlink_to(self.image_cache_dir / source_filename)
                            assert image_symlink.exists(), f"{image_symlink} is a broken symlink"

                    elif self.task == self.TASK.CLASSIFY:
                        split_dir = self.output_dir / split_name
                        source_image_path = self.image_cache_dir / source_filename
                        self._to_classify_label(category, pixel_location, source_image_path, split_dir)
            except Exception as e:
                logger.warning(f"Error making yolo annotation for entity_id: {grp.name} -- {e}")

        for split_name in dataset.data_files_df.split.unique():

            ddf = dataset.data_files_df[dataset.data_files_df.split == split_name]
            split_ids = ddf.data_file_id.unique()
            adf = dataset.annotations_df[dataset.annotations_df.data_file_id.isin(split_ids)]
            adf = adf.drop_duplicates(
                subset=["data_file_id", "entity_id", "attribute_id", "value"], keep="first"
            )

            pix_df = adf[adf.attribute_id == str(PIXEL_LOCATION_ATTRIBUTE_UUID)]
            cat_df = adf[
                (adf.attribute_id == str(self.category_attribute_id))
                & (adf.value.isin(categories_of_interest))
            ]

            pix_cat_df = pd.concat([pix_df, cat_df])
            # Add width, height and split columns annotations_df
            df = pd.merge(
                pix_cat_df,
                ddf[["data_file_id", "width", "height", "split", "filename"]],
                on="data_file_id",
                how="left",
            )

            # Ensure all attribute_id are str, sometimes they can be UUID or LabeledUUID
            df.loc[:, "attribute_id"] = df.attribute_id.map(lambda a: str(a))

            # Because entity_id can be the same from frame-to-frame as an entity
            # is tracked we need to pivot the DataFrame ('entity_id', 'data_file_id')
            # this assumes the same entity cannot appear twice in a single image.
            pivoted_df = df.pivot(index=["entity_id", "data_file_id"], columns="attribute_id", values="value")

            # Reset the index to make 'id' a column again
            pivoted_df = pivoted_df.reset_index()

            # Fill missing values with None
            pivoted_df = pivoted_df.where(pd.notnull(pivoted_df), None)

            pivoted_df = pivoted_df.merge(
                df[
                    ["data_file_id", "width", "height", "split", "filename"]
                ].drop_duplicates(),  # Select only needed columns and remove duplicates
                on="data_file_id",  # Merge key
                how="left",  # Keep all rows from pivoted_df
            )
            pivoted_df.groupby("data_file_id").apply(to_yolo_annotation)

        if self.task == self.TASK.CLASSIFY:
            images_dir = self.output_dir
        else:
            images_dir = self.output_dir / "images"

        config_dict = self.generate_data_config_dict(
            self.output_dir,
            images_dir / "train",
            images_dir / "val",
            images_dir / "test" if "test" in unique_splits else None,
            categories_of_interest,
        )
        with (self.output_dir / "data.yaml").open("w") as f:
            yaml.dump(config_dict, f)



#!/usr/bin/env python3
#
# Aiko Service: Command Line Interface
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Usage
# ~~~~~
# aiko <pipeline_definition_pathname> [--help]
#
# Description
# ~~~~~~~~~~~
# The Pipeline CLI is designed to slip ontop of existing
# pipeline definitions. I learn best with an exaple:
#
# pipeline_definition = [
#     {   "name": "VideoReadFile", "module": "aiko_services.media.video_id",
#         "successors": ["VideoWriteFile"],
#         "parameters": {
#             "video_pathname": "my-video.mp4,
#         }
#     },
#     {   "name": "VideoWriteFile", "module": "aiko_services.media.video_id",
#         "parameters": {
#             "output_pathname": "out-video.mp4,
#         }
#     },
# ]
#
# Out-of-the-box the CLI wrapper will treat everything as an optional
# parameter. To do this it infers some attributes for convenience. To
# override them create another paramerter with the same name appended
# with "_cli" and a dict of updated values.
#
#   "name": Overrides the cli parameter names
#     default: infered from component Name and parameter name
#     eg: --video-read-file-video-pathname
#     override: {"name": "-v --video-path"}
#     notes: separate short "-x" and long "--xxx" form with a space " "
#            can provide one or the other.
#
#   "required": Makes the parameter a required param from the cli
#      default: False
#      override: {"required": True}
#
#   "help": Help string
#     default: infered from component Name and parameter name
#     eg: Overrides VideoReadFile.video_pathname
#     override: {"help": "some help string"}
#
#   "hidden": Hide from CLI
#      default: False
#      override: {"hidden": True}
#
# Full Example:

# pipeline_definition = [
#     {   "name": "SomeComponent", "module": "some.module",
#         "successors": ["NextComponent"],
#         "parameters": {
#             "some_param": "default_value.mp4",
#             "some_param_cli": {
#               "name": "-v --video",
#               "help": "Better help string",
#             },
#
#             "hidden_param": "another_value.txt",
#             "hidden_param_cli": {"hidden": True}
#         }
#     },
# ]
#
# ToDo:
#   - Move --show to shim so Click doesn't complain about required params

import click
import re
import sys
import yaml
import json

from aiko_services.main import *
from aiko_services.main.utilities import *

__all__ = []

MATCH_CAMEL_CASE = re.compile(r"(?<!^)(?=[A-Z])")
DEFAULT_PIPELINE_NAME = "pipeline_definition"
DEFAULT_PIPELINE_FRAME_RATE = 0.05 # 20 FPS, 0 for flat-out!
SEP = "_SEP_"
pipeline_definition = []


def to_snake_case(val):
  return MATCH_CAMEL_CASE.sub("_", val).lower()

def cli_shim():
    """
    Intercept the cli args and determine if a path to a pipeline definition
    was passed. If so, pull it out and load it before Click gets it grubby
    little hands on it.
    """
    if len(sys.argv) >= 2 and sys.argv[1].endswith((".py", ".yaml", ".yml", ".json")):
        pipeline_definition_pathname = sys.argv[1]
        pipeline_definition, state_machine_model = load_pipeline_definition_2020(pipeline_definition_pathname)
        sys.argv.remove(pipeline_definition_pathname)
    else:
        pipeline_definition = []
        sys.argv = ["aiko", "--help"]

    return pipeline_definition


def options_from_pipeline_def(pipeline_definition):
    """
    Cli with default:

      param_name: <param> # Pipeline default
      param_name_cli: {
        "hidden": True,
        "name": # Optionally say how flags appear on cli. ie "-o --output-path"
        "required": # True|False. Default: False
        "help":      # Optional help string

        ""
        }


    """
    def infer_help(component_name, param_name):
        help_str = f"Overrides {component_name}.{param_name}"
        #if value is not None:
        #    help_str += f" [{value}]"
        return help_str

    def infer_flag(component_name, param_name):
        snake_name = to_snake_case(component_name)
        return f"--{snake_name}-{param_name}".replace("_","-").replace(" ", "-")

    def validate_attributes(attributes, component_name, param_name):
        valid_attrs = {
            "required": bool,
            "name": str,
            "help": str,
            "hidden": bool,
        }
        for k,v in attributes.items():
            if k not in valid_attrs:
                ks = list(valid_attrs.keys())
                msg = (
                    f"Invalid cli attribute {component_name}.{param_name}. "
                    f"Got: {k}. Valid attributes are {ks}.")
                raise ValueError(msg)

    def validate_value_type(value, attributes, component_name, param_name):
        if not isinstance(value, attributes["type"]):
            t = attributes["type"]
            msg = (
                f"Type missmatch in {component_name}.{param_name} "
                f"type of {t} specified, but got {type(value)}"
            )
            raise ValueError(msg)


    def decorator(f):
        for ele in reversed([e for e in pipeline_definition if "parameters" in e]):
            component_name = ele["name"]

            # Required cli params
            params = {k:v for k,v in ele["parameters"].items() if not k.endswith("_cli")}
            cli_attrs = {k:v for k,v in ele["parameters"].items() if k.endswith("_cli")}
            for param_name, value in params.items():
                attributes = cli_attrs.pop(f"{param_name}_cli", {})
                if attributes.get("hidden", False):
                    continue
                validate_attributes(attributes, component_name, param_name)

                # Fill in missing attributes as needed
                if not attributes.get("required", False):
                    attributes["default"] = value
                    attributes["show_default"] = True


                attributes["type"] = type(value)
                if "help" not in attributes:
                    attributes["help"] = infer_help(component_name, param_name)

                infered_flag = infer_flag(component_name, param_name)
                flags = attributes.get("name", infered_flag).split()

                validate_value_type(value, attributes, component_name, param_name)
                variable_name = f"{component_name}{SEP}{param_name}"
                flags += [variable_name]
                click.option(*flags, **attributes)(f)
        return f
    return decorator


def clean_cli_params(pipeline_definition):
    for component in [e for e in pipeline_definition if "parameters" in e]:
        param_names = list(component["parameters"].keys())
        for param_name in param_names:
            if param_name.endswith("_cli"):
                component["parameters"].pop(param_name)
    return pipeline_definition

pipeline_definition = cli_shim()

@click.command("main", help=(
  "Load Pipeline Definition, build cli, override parameters if needed, run "
  "pipeline. Like a boss!\n\nUsage:\n\n"

  "  aiko <pipeline_definition.py> --help"))
@options_from_pipeline_def(pipeline_definition)
@click.option("--pipeline-frame-rate", "-fps", type=int,
  required=False, default=DEFAULT_PIPELINE_FRAME_RATE,
  help=f"Rate at which to run the pipeline over frames [{DEFAULT_PIPELINE_FRAME_RATE}]")
@click.option("--show", is_flag=True,
  help="Only print the pipeline, dont run it.")
@click.option("--dump", type=str, default=None,
  help="Save the file to yaml or json")

def main(**kwargs):
    dump = kwargs["dump"]
    if dump is not None:
        to_dump = {"pipeline_definition": pipeline_definition}
        if dump.endswith((".yaml", ".yml")):
            with open(dump, "w") as f:
                yaml.dump(to_dump, f)
        elif dump.endswith(".json"):
            with open(dump, "w") as f:
                json.dump(to_dump, f, indent=2)
        else:
            raise ValueError(f"Invalid file type: got {dump}")
        return 0

    _pipeline_def = clean_cli_params(pipeline_definition)
    pipeline = Pipeline_2020(_pipeline_def, kwargs["pipeline_frame_rate"])

    for k,v in kwargs.items():
        if SEP in k:
            node_name, param_name = k.split(SEP)
            pipeline.update_node_parameter(node_name, param_name, v)

    if kwargs["show"]:
        d = []
        for node in _pipeline_def:
            node_name = node["name"]
            info = pipeline.get_node(node_name)
            d.append({
                "name": node_name,
                "module": info["module"],
                "successors": info["successors"],
                "parameters": info["parameters"]
            })
        print(yaml.dump(d))
    else:
        pipeline.run(mqtt_connection_required=False)

if __name__ == "__main__":
    main()



import json
import logging
import os
import shutil
import tarfile
import tempfile
import zipfile
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Tuple, Union
from uuid import UUID

import numpy as np
import pandas as pd
import yaml
from shapely.geometry.base import BaseGeometry as ShapelyShape
from shapely.wkt import loads as wkt_loads

from highlighter.client.training_config import TrainingConfigType
from highlighter.datasets.interfaces import IReader
from highlighter.datasets.interpolation import (
    interpolate_pixel_locations_between_frames,
)

from ..client import DatasetSubmissionTypeConnection as DatasetSubmissionConnection
from ..client import (
    HLClient,
    create_data_files,
    download_file_from_s3,
    list_files_in_s3,
    multithread_graphql_file_download,
    read_object_classes,
)
from ..core import (
    OBJECT_CLASS_ATTRIBUTE_UUID,
    PIXEL_LOCATION_ATTRIBUTE_UUID,
    GQLBaseModel,
    paginate,
)
from .base_models import (
    CLOUD_FILES_INFO_KEY,
    DEFAULT_ANNOS_KEY,
    DEFAULT_DATA_FILES_KEY,
    AttributeRecord,
    ImageRecord,
    S3Files,
)

KEY_RECORDS = "records"
KEY_FILES = "files"
MANIFEST_YAML = "manifest.yaml"
AWS_S3 = "aws-s3"

__all__ = ["Dataset", "DatasetFormat", "dataset_in_cloud"]

LOG = logging.getLogger(__name__)


# Custom representer function for MyUUID
def uuid_representer(dumper, data):
    """Represent UUID|LabeledUUID as a string"""
    return dumper.represent_scalar("tag:yaml.org,2002:str", str(data))


yaml.add_representer(UUID, uuid_representer)


class DatasetFormat:
    HIGHLIGHTER_WORKFLOW = "highlighter-workflow"
    HIGHLIGHTER_DATASET = "highlighter-dataset"
    JSON = "json"
    HDF = "hdf"
    AWS_S3 = "aws-s3"
    COCO = "coco"
    HIGHLIGHTER_ASSESSMENTS = "highlighter-assessments"
    DATA_FILE_FOLDER = "torch-data_file-folder"


class TempDirAt:
    def __init__(self, path):
        self.path = Path(path)

    def __enter__(self):
        self.path.mkdir(exist_ok=False, parents=True)
        return self.path

    def __exit__(self, type, value, traceback):
        shutil.rmtree(str(self.path))


def md5sum_from_prefix(file_prefix: str):
    parts = Path(file_prefix).stem.split("_")
    if len(parts) == 1:
        # No md5sum in file_prefix
        return None
    else:
        return parts[-1]


def is_uuid(v):
    _is_uuid = False
    try:
        _v = UUID(v)
        _is_uuid = True
    except (ValueError, AttributeError):
        pass
    return _is_uuid


def is_enum_type(v):
    result = False
    if isinstance(v, UUID):
        result = str(v)
    elif is_uuid(v):
        result = v
    return result


def dataset_in_cloud(
    client: HLClient,
    dataset_id: int,
) -> Tuple[bool, Optional[str]]:
    """Discover if the dataset is in 3rd party cloud storage"""

    class DatasetInfo(GQLBaseModel):
        id: int
        location_uri: Optional[str] = None
        format: str

    if client.cloud_creds is None:
        return None, None

    result = client.dataset(
        return_type=DatasetInfo,
        id=dataset_id,
    )

    is_cloud_dataset = result.location_uri is not None
    return is_cloud_dataset, result.location_uri


def get_value_type(v):
    if isinstance(v, bool):
        return v, "boolean"

    _v = is_enum_type(v)
    if _v:
        return _v, "enum"

    if "POLYGON" in str(v):
        return v, PIXEL_LOCATION_ATTRIBUTE_UUID.label

    if isinstance(v, np.ndarray):
        return v, "numpy.ndarray"

    return v, type(v).__name__


def unpack_archive(archive_path: Path, unpack_dir: Path):
    archive_path = Path(archive_path)
    unpack_dir = Path(unpack_dir)

    ext = archive_path.suffix
    if ext == ".zip":
        with zipfile.ZipFile(str(archive_path), "r") as storage:
            storage.extractall(unpack_dir)  # nosec tarfile_unsafe_members

    elif ext in (".tar.gz", ".tar"):
        with tarfile.open(str(archive_path)) as storage:
            storage.extractall(unpack_dir, filter="data")


def download_s3_files_archives(
    client: HLClient,
    bucket_name: str,
    files_prefixes: List[Path],
    data_files_cache_dir: Path,
) -> List[Path]:
    unpacked_files_list: List[str] = []

    for files_prefix in files_prefixes:
        md5sum = md5sum_from_prefix(files_prefix)
        file_cache_path = data_files_cache_dir / Path(files_prefix).name
        file_cache_marker = data_files_cache_dir / f"CACHED_{Path(files_prefix).stem}.yaml"

        _unpacked_files_list: List[str] = []
        if file_cache_marker.exists():
            with file_cache_marker.open("r") as f:
                data = yaml.safe_load(f)

            # Pull in files that have already been downloaded
            unpacked_files_list.extend(data["unpacked_files_list"])
            message = data["message"]
            LOG.info(f"Cache files marker found, {message}")
            continue

        LOG.info(f"Downloading s3://{bucket_name}/{files_prefix}")
        download_file_from_s3(
            client,
            bucket_name,
            str(files_prefix),
            str(file_cache_path),
            md5sum=md5sum,  # <-- If None, will not perform check
        )

        # Open a temporary directory in the data_files_cache_dir to unpack the
        # files archive before moving the contents to the final destination
        with TempDirAt(data_files_cache_dir / f"tmp_{files_prefix.stem}") as tmp:
            unpack_archive(file_cache_path, tmp)

            unpacked_files_dir = list(tmp.glob("*"))
            assert len(unpacked_files_dir) == 1
            unpacked_files_dir = unpacked_files_dir[0]

            # Move files from data_files_cache_dir/files/* to data_files_cache_dir/
            for f in unpacked_files_dir.rglob("*"):
                # Only interested in moving files not dirs
                if f.is_dir():
                    continue

                # Remove dest file if it exists already
                unpacked_file_rel_path = f.relative_to(unpacked_files_dir)

                # Add to unpacked_files_list so we can return it
                _unpacked_files_list.append(str(unpacked_file_rel_path))

                # Remove dest if exists
                dest = Path(data_files_cache_dir) / unpacked_file_rel_path
                if dest.exists():
                    dest.unlink()

                # Make destination dir as needed
                dest.parent.mkdir(exist_ok=True, parents=True)

                # Move the file
                shutil.move(str(f), str(dest))

        # Remove archive file after unpacking.
        os.remove(str(file_cache_path))

        # Create a marker file to indicate the file was downloaded on
        # a given date
        with file_cache_marker.open("w") as f:
            message = f"file s3://{bucket_name}/{files_prefix} downloaded at {datetime.now()}"
            yaml.dump(
                {
                    "message": message,
                    "unpacked_files_list": _unpacked_files_list,
                },
                f,
            )

        unpacked_files_list += _unpacked_files_list
        LOG.info(f"Unpacked {len(_unpacked_files_list)} files")
    return unpacked_files_list


class Dataset:
    REQUIRED_ANNOTATIONS_DF_COLUMNS = set(AttributeRecord.model_fields.keys())
    REQUIRED_DATA_FILES_DF_COLUMNS = set(ImageRecord.model_fields.keys())

    def __init__(
        self,
        dataset_id: int = None,
        annotations_df=None,
        data_files_df=None,
        cloud_files_info: Union[S3Files, List[S3Files]] = None,
        attribute_records: Optional[List[AttributeRecord]] = None,
        data_file_records: Optional[List[ImageRecord]] = None,
    ):
        """

        Params:
            annotations_df: Pandas DataFrame with rows representing AttributeRecord

            data_files_df: Pandas DataFrame with rows representing ImageRecord

            cloud_files_info: Information needed to load download files from a
                              cloud services. At this point we only support S3.
                              See S3Files for more info
            attribute_records: List of AttributeRecords to initialize Dataset with, must
                               include data_file_records too

            data_file_records: List of ImageRecords to initialize Dataset with, must
                               include attribute_records too

        """
        self.dataset_id = dataset_id

        if (annotations_df is not None) and (data_files_df is not None):
            self.annotations_df = annotations_df
            self.data_files_df = data_files_df
        elif attribute_records is not None:
            self.annotations_df = pd.DataFrame([r.to_df_record() for r in attribute_records])
            if data_file_records is not None:
                self.data_files_df = pd.DataFrame([r.model_dump() for r in data_file_records])
            else:
                self.data_files_df = pd.DataFrame()

        # Due to a bug in HLWeb, when tracks are joined sometimes the annotation at the joining
        # frame can be duplicated. Here we drop duplicates and keep
        # most recent ("last" in the DataFrame) annotation.
        if ("frame_id" not in self.annotations_df) and (
            "frame_id" in getattr(self.annotations_df.iloc[0], "extra_fields", {})
        ):
            if self.annotations_df.iloc[0].extra_fields.get("frame_id", None) is not None:
                self.annotations_df["frame_id"] = self.annotations_df.extra_fields.apply(
                    lambda x: x["frame_id"]
                )
        if "frame_id" in self.annotations_df:
            if self.annotations_df.frame_id.hasnans:
                raise ValueError("Unable to deduplicate on frame_id if some are NaN")
            self.annotations_df = self.annotations_df.drop_duplicates(
                subset=["frame_id", "data_file_id", "entity_id", "attribute_id"], keep="last"
            )

        if cloud_files_info is None:
            cloud_files_info = []
        if isinstance(cloud_files_info, (S3Files, dict)):
            cloud_files_info = [cloud_files_info]

        self.cloud_files_info = []
        for c in cloud_files_info:
            if isinstance(c, dict):
                self.cloud_files_info.append(S3Files(**c))
            elif isinstance(c, S3Files):
                self.cloud_files_info.append(c)
            else:
                raise ValueError(f"Expected dict or S3Files object got: {c}")

    @property
    def attributes_df(self):
        """For when I'm lazy and don't want to type annotations_df"""
        return self.annotations_df

    def get_unique_categories(self, attribute_id: UUID) -> List:
        return self.attributes_df[self.attributes_df.attribute_id == attribute_id].value.unique().tolist()

    @classmethod
    def get_reader(cls, dataset_format: str):
        readers = {
            DatasetFormat.JSON: cls.read_json,
            DatasetFormat.HDF: cls.read_hdf,
            DatasetFormat.AWS_S3: cls.read_s3,
            DatasetFormat.HIGHLIGHTER_DATASET: cls.read_highlighter_dataset_assessments,
            DatasetFormat.HIGHLIGHTER_WORKFLOW: cls.read_highlighter_workflow_assessments,
            DatasetFormat.HIGHLIGHTER_ASSESSMENTS: cls.read_assessments_gen,
            DatasetFormat.COCO: cls.read_coco,
            DatasetFormat.DATA_FILE_FOLDER: cls.read_data_file_folder,
        }

        if dataset_format not in readers:
            raise ValueError(f"Invalid dataset format: '{dataset_format}' for reader")

        return readers[dataset_format]

    @classmethod
    def _read_cached_dataset(
        cls,
        dataset_path: Path,
        **reader_kwargs,
    ) -> Tuple[pd.DataFrame, pd.DataFrame, Optional[pd.DataFrame], Optional[S3Files]]:
        # Drop the '.' from the suffix, ie: '.json' -> 'json'
        dataset_format = dataset_path.suffix[1:]
        reader = cls.get_reader(dataset_format)

        return reader(path=dataset_path, **reader_kwargs)

    @classmethod
    def read_coco(cls, annotations_file: Path, bbox_only: bool = False):
        from . import get_reader

        coco_reader = get_reader("coco")(annotations_file, bbox_only=bbox_only)
        return cls.load_from_reader(coco_reader)

    @classmethod
    def read_from(
        cls,
        dataset_format: Union[str, DatasetFormat],
        data_files_cache_dir: Union[str, Path] = None,
        **reader_kwargs,
    ):
        # TODO: All readers must deal with their own caching.

        reader = cls.get_reader(dataset_format)

        ds = reader(**reader_kwargs)

        """It is desirable to add the dataset_id to the annotations_df so a
        dev can easily inspect the data and know where it came from. This is
        especially useful when working with datasets that consist of 2 or more
        other smaller datasets.
        """
        dataset_id = reader_kwargs.get("dataset_id", None)
        if dataset_id is not None:
            ds.annotations_df["dataset_id"] = dataset_id

        """If data_files_cache_dir is provided we download the data_file files
        associated with the datasets.
        """
        if data_files_cache_dir is not None:
            data_files_cache_dir = Path(data_files_cache_dir)
            client = reader_kwargs.get("client")
            assert isinstance(client, HLClient), (
                "if `data_files_cache_dir` is set you must provide a valid "
                f"`client` in `reader_kwargs`, got: '{client}'"
            )

            cls.download_dataset_files(
                client,
                data_files_cache_dir,
                data_files_df=ds.data_files_df,
                cloud_files_info=ds.cloud_files_info,
            )

        return ds

    def append(
        self, datasets: List["Dataset"], drop_duplicates_keep: Optional[Literal[True, "first", "last"]] = None
    ):
        if not isinstance(datasets, (tuple, list)):
            datasets = [datasets]

        data_file_dfs = [d.data_files_df for d in datasets]
        annotations_dfs = [d.annotations_df for d in datasets]

        adf = pd.concat([self.annotations_df] + annotations_dfs, ignore_index=True)
        ddf = pd.concat([self.data_files_df] + data_file_dfs, ignore_index=True)

        # When duduplicating we need to convert the extra_fields column to something
        # that can be compared. So we conver them to str temporarly, do
        # the dedupe then add the originals back into the deduped data frames.
        if drop_duplicates_keep is not None:
            if hasattr(adf, "extra_fields"):
                adf_extra = adf.extra_fields
                adf["extra_fields"] = adf.extra_fields.apply(lambda x: str(x))
                adf_dup = adf.duplicated(keep=drop_duplicates_keep).values
                adf = adf.loc[np.logical_not(adf_dup)]
                adf["extra_fields"] = adf_extra
            else:
                adf = adf.drop_duplicates(keep=drop_duplicates_keep)

            if hasattr(ddf, "extra_fields"):
                ddf_extra = ddf.extra_fields
                ddf["extra_fields"] = ddf.extra_fields.apply(lambda x: str(x))
                ddf_dup = ddf.duplicated(keep=drop_duplicates_keep).values
                ddf = ddf.loc[np.logical_not(ddf_dup)]
                ddf["extra_fields"] = ddf_extra
            else:
                ddf = ddf.drop_duplicates(keep=drop_duplicates_keep)

        self.annotations_df = adf
        self.data_files_df = ddf

        for dataset in datasets:
            self.cloud_files_info.extend(dataset.cloud_files_info)

    def apply_split(self, dataset_splitter: "DatasetSplitter"):
        (self.data_files_df, self.annotations_df) = dataset_splitter.split(self)

    def get_stats(self, split=None, uuid_to_name=None):
        stats_dict = dict(data_files=dict(), attributes=[])

        if split is not None:
            data_files_df = self.data_files_df[self.data_files_df.split == split]
            if data_files_df.shape[0] == 0:
                unique_splits = self.data_files_df.split.unique()
                raise ValueError(f"No split '{split}' found in dataset. " f"Expected one of; {unique_splits}")

            data_file_ids = data_files_df.data_file_id
            annotations_df = self.annotations_df[self.annotations_df.data_file_id.isin(data_file_ids)]
        else:
            data_files_df = self.data_files_df
            annotations_df = self.annotations_df

        stats_dict["data_files"]["count"] = data_files_df.shape[0]

        for attr_id in annotations_df.attribute_id.unique():
            attr_df = annotations_df[annotations_df.attribute_id == attr_id]

            attr_record = dict(
                id=attr_id,
                name=attr_df.iloc[0].attribute_name,
                count=attr_df.shape[0],
            )

            value, value_type = get_value_type(attr_df.iloc[0].value)
            attr_record["value_type"] = value_type
            attr_record["total"] = attr_df.shape[0]
            if value_type == "enum":
                attr_record["member_counts"] = dict()
                for enum_id in attr_df.value.unique():
                    count_dict = {"count": attr_df[attr_df.value == enum_id].shape[0]}

                    enum_name = uuid_to_name.get(enum_id, None)
                    if enum_name is not None:
                        count_dict["name"] = enum_name

                    attr_record["member_counts"][enum_id] = count_dict

            elif value_type == "boolean":
                attr_record["member_counts"] = dict()
                for val in [True, False]:
                    attr_record["member_counts"][str(val)] = attr_df[attr_df.value == val].shape[0]
            stats_dict["attributes"].append(attr_record)
        return stats_dict

    def _get_uuid_to_name_lookup(self, client):
        """Get uuid for all enum values"""

        # ToDo: Also get attribute names when gql allows it.

        mask = self.annotations_df.value.apply(is_uuid)
        uuid_values = self.annotations_df[mask].value.unique()
        object_classes = read_object_classes(client, uuid=uuid_values.tolist())
        return {str(o.uuid): o.name for o in object_classes}

    def publish_to_highlighter(
        self,
        client: HLClient,
        dataset_name: str,
        dataset_description_fields: List[Tuple[str, str]] = [],
        split_fracs: Dict[str, int] = {},
    ):
        uuid_to_name = self._get_uuid_to_name_lookup(client)

        id_split_name_url = []
        ids = []
        # Loop over the unique splits and create a dataset in
        # highlighter without populating it. We do this so we
        # can get the dataset_ids upfront so we can generate
        # urls to the various splits.
        for split_name in self.data_files_df.split.unique():
            split_frac = split_fracs.get(split_name, None)
            if split_frac is not None:
                split_str = f"{split_name}-{split_frac}"
            else:
                split_str = split_name

            name = "_".join(
                [
                    f"{datetime.now().strftime('%Y-%m-%d')}",
                    f"{dataset_name}",
                    f"{split_str}",
                ]
            )

            class DatasetType(GQLBaseModel):
                id: int

            class CreateDatasetPayload(GQLBaseModel):
                dataset: Optional[DatasetType] = None
                errors: list

            class DatasetPayload(GQLBaseModel):
                dataset: Optional[DatasetType] = None
                errors: Optional[list] = None

            class SubsAndHashes(GQLBaseModel):
                id: int
                hash_signature: str

            response = client.createDataset(
                return_type=CreateDatasetPayload,
                name=name,
                description=name,
            )

            if len(response.errors) > 0:
                raise ValueError(response.errors)

            id = response.dataset.id
            url = client.endpoint_url.replace("graphql", f"datasets/{id}")
            id_split_name_url.append((id, split_name, name, url))
            ids.append(id)
            print(f"Created dataset: {id}")

        def fix_underscores(n):
            return n.replace("_", "\\_")

        dataset_markdown_links = [
            f"[{id}]({url}): {fix_underscores(name)}" for id, _, name, url in id_split_name_url
        ]
        stats_dict = self.get_stats(uuid_to_name=uuid_to_name)
        global_stats_str = yaml.dump(stats_dict)

        df = self.data_files_df
        for idx, (id, split_name, name, url) in enumerate(id_split_name_url):
            records = df.loc[df.split == split_name, ["assessment_id", "hash_signature"]].to_dict("records")

            subs_and_hashes = [
                SubsAndHashes(
                    id=r["assessment_id"],
                    hash_signature=r["hash_signature"],
                ).gql_dict()
                for r in records
            ]

            response = client.populateDataset(
                return_type=DatasetPayload,
                datasetId=id,
                submissionIdsAndHashes=subs_and_hashes,
            )

            if response.errors:

                class DatasetPopulateError(Exception):
                    pass

                raise DatasetPopulateError("\n".join(response.errors))

            dataset_description = [
                f"# {dataset_name.replace('-', ' ').replace('_', ' ').title()}\n",
                f"**Dataset {id}**",
            ]

            if len(id_split_name_url) > 1:
                # Make list of links
                # Make current split bold but remove url link
                def make_bold_remove_url(l):
                    return f'**{l[1:].split("]")[0]}**'

                links = [
                    f"  - {make_bold_remove_url(l)}: You are here 😀" if i == idx else f"  - {l}"
                    for i, l in enumerate(dataset_markdown_links)
                ]

                links_str = "\n".join(links)
                _dataset_description_fields = [("Related Splits", links_str)] + dataset_description_fields
            else:
                _dataset_description_fields = dataset_description_fields

            dataset_description.extend(
                [f"## {heading}\n\n{value}\n" for heading, value in _dataset_description_fields]
            )

            split_stats_str = yaml.dump(
                self.get_stats(
                    split=split_name,
                    uuid_to_name=uuid_to_name,
                )
            )

            dataset_description.extend(
                [
                    f"## {split_name.title()} Split Stats\n\n<pre>{split_stats_str}</pre> \n",
                ]
            )

            dataset_description.extend(
                [
                    f"## Golbal Stats\n\n<pre>{global_stats_str}</pre> \n",
                ]
            )

            dataset_description_str = "\n".join(dataset_description)

            response = client.updateDataset(
                return_type=DatasetPayload,
                id=id,
                description=dataset_description_str,
            )

            if len(response.errors) > 0:
                raise ValueError(response.errors)

            print(f"Populated Dataset: {id} with {len(subs_and_hashes)} assessments")

            response = client.lockDataset(
                return_type=DatasetPayload,
                datasetId=id,
            )
            print(f"Locked Dataset: {id}")
            print(f"See dataset at: {url}")
        return ids

    @classmethod
    def combine(cls, datasets: List["Dataset"]):
        if len(datasets) == 1:
            return datasets[0]

        if len(datasets) == 0:
            raise ValueError("Expected list of Dataset objects, got []")

        base_ds = datasets[0]
        base_ds.append(datasets[1:])

        return base_ds

    @classmethod
    def download_dataset_files(
        cls,
        client: HLClient,
        data_files_cache_dir: Path,
        data_files_df: pd.DataFrame = None,
        cloud_files_info: List[S3Files] = None,
        **kwargs,
    ):
        data_files_cache_dir = Path(data_files_cache_dir)

        existing_file_paths: List[Path] = []
        if cloud_files_info is not None:
            for info in cloud_files_info:
                files_prefixes = [Path(info.prefix) / file for file in info.files]

                # paths relative to data_files_cache_dir
                existing_file_paths = download_s3_files_archives(
                    client,
                    info.bucket_name,
                    files_prefixes,
                    data_files_cache_dir,
                )

        if data_files_df is not None:
            existing_file_strs: List[str] = [str(p) for p in existing_file_paths]

            data_files_to_download = data_files_df[
                ~data_files_df.filename.isin(existing_file_strs)
            ].data_file_id.unique()
            multithread_graphql_file_download(
                client,
                list(data_files_to_download),
                data_files_cache_dir,
                **kwargs,
            )

    @classmethod
    def read_highlighter_workflow_assessments(
        cls,
        *,
        client: HLClient,
        queryArgs: Dict,
        **kwargs,
    ):
        """Instantiate a Dataset from a Highlighter workflow assessments.

        You can provide dict of queryArgs that will be used to compile a
        GraphQL query. The resulting assessments will populate the Dataset

        If you need to download accompanying data_files you can
        either use the generic `Dataset.read` classmethod or use
        `Dataset.download_dataset_files`
        """
        from highlighter.client import get_latest_assessments_gen

        assessments_gen = get_latest_assessments_gen(
            client,
            **queryArgs,
        )

        return cls.read_assessments_gen(assessments_gen=assessments_gen)

    @classmethod
    def read_highlighter_dataset_assessments(
        cls,
        client: HLClient,
        dataset_id: int,
        datasets_cache_dir: Optional[Path] = None,
        page_size=100,
        **kwargs,
    ):
        """Check for cached dataset"""
        if datasets_cache_dir is not None:
            datasets_cache_dir = Path(datasets_cache_dir)
            dataset_cache_path = datasets_cache_dir / f"records_{dataset_id}.json"
            if dataset_cache_path.exists():
                return cls.read_json(path=dataset_cache_path)
        else:
            dataset_cache_path = None

        _assessments_gen = paginate(
            client.datasetSubmissionConnection,
            DatasetSubmissionConnection,
            page_size=page_size,
            datasetId=dataset_id,
        )

        # datasetSubmissionConnection nests the SubmissionType
        # inside the node object as opposed to the SubmissionType
        # being the node object. So we unpack it here so to
        # adhear to a consistent interface
        assessments_gen = (node.submission for node in _assessments_gen)

        ds = cls.read_assessments_gen(assessments_gen=assessments_gen)

        """If we can, cache the dataset locally
        """
        if dataset_cache_path is not None:
            ds.write_json(dataset_cache_path)

        return ds

    @classmethod
    def read_assessments_gen(
        cls,
        assessments_gen,
    ):
        """Load data_files_df, annotations_df from a Highlighter assessments
        generator. Returns these as a tuple to be used by the generic
        `Dataset.read` classmethod
        """
        from highlighter.datasets.formats.highlighter.reader import (
            HighlighterAssessmentsReader,
        )

        reader = HighlighterAssessmentsReader(assessments_gen)
        return cls.load_from_reader(reader)

    def download_files_from_datasource(
        self,
        client: HLClient,
        data_files_cache_dir: Path,
        **kwargs,
    ):
        data_files_cache_dir = Path(data_files_cache_dir)
        multithread_graphql_file_download(
            client,
            list(self.data_files_df.data_file_id.unique()),
            data_files_cache_dir,
            **kwargs,
        )

    @classmethod
    def read_hdf(
        cls,
        path: Path,
        data_files_key: Optional[str] = DEFAULT_DATA_FILES_KEY,
        annotations_key: Optional[str] = DEFAULT_ANNOS_KEY,
        **kwargs,
    ):
        """Instantiate a Dataset from a local .hdf file

        If you need to download accompanying data_files you can
        either use the generic `Dataset.read` classmethod or use
        `Dataset.download_dataset_files`
        """
        path = Path(path)

        annotations_df = pd.read_hdf(path, key=annotations_key)
        data_files_df = pd.read_hdf(path, key=data_files_key)

        try:
            cloud_files_info = pd.read_hdf(path, key=CLOUD_FILES_INFO_KEY)
            cloud_files_info = [S3Files.safe_load(**info) for info in cloud_files_info.to_dict("records")]
        except KeyError:
            # cloud files key is optional
            cloud_files_info = None

        return cls(
            data_files_df=data_files_df,
            annotations_df=annotations_df,
            cloud_files_info=cloud_files_info,
        )

    def write_json(
        self,
        path: Path,
        data_files_key: Optional[str] = DEFAULT_DATA_FILES_KEY,
        annotations_key: Optional[str] = DEFAULT_ANNOS_KEY,
    ):
        payload = {}
        if len(self.cloud_files_info) > 0:
            payload[CLOUD_FILES_INFO_KEY] = [c.dict() for c in self.cloud_files_info]

        payload[annotations_key] = self.annotations_df.to_dict("records")
        payload[data_files_key] = self.data_files_df.to_dict("records")

        class ShapelyGeometryEncoder(json.JSONEncoder):
            def default(self, obj: Any) -> Any:
                if isinstance(obj, ShapelyShape):
                    return obj.wkt  # Convert BaseGeometry instances to WKT
                # Let the base class default method raise the TypeError
                return json.JSONEncoder.default(self, obj)

        path = Path(path)
        with path.open("w") as f:
            json.dump(payload, f, cls=ShapelyGeometryEncoder)

    @classmethod
    def read_data_file_folder(
        cls,
        *,
        path: Path,
        attribute_id: str = str(OBJECT_CLASS_ATTRIBUTE_UUID),
        attribute_name: str = OBJECT_CLASS_ATTRIBUTE_UUID.label,
    ):
        from highlighter.datasets.formats.torch_image_folder.reader import (
            TorchImageFolderReader,
        )

        reader = TorchImageFolderReader(path, attribute_id=attribute_id, attribute_name=attribute_name)
        return cls.load_from_reader(reader)

    @classmethod
    def load_from_reader(cls, reader: IReader):
        data_file_records, attribute_records = reader.read()

        if len(data_file_records) == 0:
            raise ValueError(
                (
                    f"Could not populate Dataset object from "
                    f"{reader}. This could be because the "
                    f"HLClient.endpoint_url is incorrect, or, if "
                    "The dataset is stored as a 'cloud dataset' maybe the cloud credenials "
                    "are incorrect/missing"
                )
            )

        annotations_df = pd.DataFrame([r.to_df_record() for r in attribute_records])
        data_files_df = pd.DataFrame([r.model_dump() for r in data_file_records])
        return cls(data_files_df=data_files_df, annotations_df=annotations_df)

    @classmethod
    def read_json(
        cls,
        path: Path,
        data_files_key: Optional[str] = DEFAULT_DATA_FILES_KEY,
        annotations_key: Optional[str] = DEFAULT_ANNOS_KEY,
        **kwargs,
    ):
        """Instantiate a Dataset from a local .json file

        If you need to download accompanying data_files you can
        either use the generic `Dataset.read` classmethod or use
        `Dataset.download_dataset_files`
        """
        # Make sure path is a Path object
        path = Path(path)

        class ShapelyGeometryDecoder(json.JSONDecoder):
            def __init__(self, *args, **kwargs):
                super().__init__(object_hook=self.dict_to_object, *args, **kwargs)
                self.WKT_INDICATORS = ("POLYGON", "MULTIPOLYGON", "POINT", "LINESTRING")

            def dict_to_object(self, d: dict) -> dict:
                if (
                    ("value" in d)
                    and isinstance(d["value"], str)
                    and (d["value"].split(" ")[0] in self.WKT_INDICATORS)
                ):
                    d["value"] = wkt_loads(d["value"])
                return d

        with path.open("r") as f:
            data = json.load(f, cls=ShapelyGeometryDecoder)

        if annotations_key in data:
            attr_list = data[annotations_key]
        else:
            attr_list = data["attributes"]

        annotations_df = pd.DataFrame(attr_list)

        data_files_df = pd.DataFrame(data[data_files_key])

        cloud_files_info = data.get(CLOUD_FILES_INFO_KEY, None)

        if isinstance(cloud_files_info, dict):
            """Some older instances of cloud_files_info have a single dict
            not a list as is standard now
            """
            cloud_files_info = [cloud_files_info]

        assert (cloud_files_info is None) or isinstance(cloud_files_info, list)

        if isinstance(cloud_files_info, list):
            cloud_files_info = [S3Files(**info) for info in cloud_files_info]

        return cls(
            data_files_df=data_files_df,
            annotations_df=annotations_df,
            cloud_files_info=cloud_files_info,
        )

    @classmethod
    def read_s3(
        cls,
        client: HLClient,
        dataset_id: int,
        datasets_cache_dir: Path = None,
        data_files_key: Optional[str] = DEFAULT_DATA_FILES_KEY,
        annotations_key: Optional[str] = DEFAULT_ANNOS_KEY,
        **kwargs,
    ):
        """
        Will attempt to download dataset records from s3 and instantiate
        a Dataset object. Optioally will download files associated with records.

        NOTE: we're in the process of renaming `annotations` to `attributes` and
              `data_files` to `files`. So keep that in mind when reading this doc
              string.

        Highlighter Datasets are stored a directory with the `training_run_id`
        as the name. This dir contains two types of files; `records` and `files`.

        s3://my-bucket/datasets/123/       <-- `dataset_id` 123
            records_<md5sum>.<json | hdf>  <-- contains EAVT information
            files_<md5sum>.tar.gz          <-- one or more file archives contain
                                            the files references in `records`
            ...

        records: Is a `.json` or `.hdf` file containing Entity Attribute Value Type
                 information. This information is loaded into the Dataset's
                 `data_files_df` and `annotations_df`. For more information on the
                 underlying Dataframe's and their schema see the doc string
                 for the Dataset class.

        files: A `.tar.gz|.tar|.zip` archive file containing the files refered
               to in the `data_files_df`. If there is a large number of files these
               archives can be broken up into smaller chunks each with a unique
               md5sum. When unpacked, each archive must contain a singel directory
               names 'files' that contains the files refered to in `data_files_df`.
               NOTE: `data_files_df.filename` should NOT include 'files/' as this
               dir will be dropped when unpacking the archive.

        Params:
            bucket_name: name of s3 bucket. (above example = 'my-bucket')

            dataset_id: id of dataset to download  (above example = 123)

            client: a HLClient with the correct cloud credenials

            prefix: Relitave path to datasets dir (above example = 'datasets/123')

            files_cache: If supplied will download files archive from s3 and
                         unpack here
        """

        """Check for cached dataset
        """
        if datasets_cache_dir is not None:
            datasets_cache_dir = Path(datasets_cache_dir)

            dataset_cache_path = datasets_cache_dir / f"records_{dataset_id}.json"
            if dataset_cache_path.exists():
                return cls.read_json(path=dataset_cache_path)

        """No cached dataset found, read from s3
        """

        class DatasetInfo(GQLBaseModel):
            id: int
            location_uri: Optional[str] = None
            format: str

        if client.cloud_creds is None:
            raise ValueError("Cannot read an s3 dataset without cloud_creds")

        result = client.dataset(
            return_type=DatasetInfo,
            id=dataset_id,
        )

        location_uri = result.location_uri

        assert location_uri.startswith("s3://")

        # S3 bucket_name is the first part in the uri after the s3://
        bucket_name = location_uri[5:].split("/")[0]

        prefix = "/".join(location_uri[5:].split("/")[1:])

        # the aws s3 cli treats prefixes with a trailing /
        # differenly to those without. It seems Boto3 is not
        # as tempormental, but just to be consistent.
        if prefix.endswith("/"):
            prefix = prefix[:-1]

        s3_contents = list_files_in_s3(
            client,
            bucket_name,
            prefix,
        )

        # Download MANIFEST_YAML that contains list of files in the dataset
        manifest_file_prefix = [x for x in s3_contents if Path(x).name == MANIFEST_YAML]
        assert len(manifest_file_prefix) == 1
        manifest_file_prefix = manifest_file_prefix[0]
        with tempfile.TemporaryDirectory() as tmp:
            tmp_manifes_path = Path(tmp) / MANIFEST_YAML
            download_file_from_s3(
                client,
                bucket_name,
                str(manifest_file_prefix),
                str(tmp_manifes_path),
            )
            with tmp_manifes_path.open("r") as f:
                manifest = yaml.safe_load(f)

        with tempfile.TemporaryDirectory() as tmp:
            records_filename = manifest[KEY_RECORDS][0]
            records_md5sum = md5sum_from_prefix(records_filename)
            tmp_path = Path(tmp) / records_filename

            records_file_prefix = f"{prefix}/{records_filename}"
            download_file_from_s3(
                client,
                bucket_name,
                records_file_prefix,
                str(tmp_path),
                md5sum=records_md5sum,  # <-- If None, will not perform check
            )

            ext = tmp_path.suffix[1:]
            reader = cls.get_reader(ext)
            ds = reader(
                path=tmp_path,
                data_files_key=data_files_key,
                annotations_key=annotations_key,
            )

        """If we can, cache the dataset locally
        """
        if datasets_cache_dir is not None:
            ds.write_json(dataset_cache_path)

        return ds

    def write_hdf(
        self,
        path: Path,
        data_files_key: Optional[str] = DEFAULT_DATA_FILES_KEY,
        annotations_key: Optional[str] = DEFAULT_ANNOS_KEY,
    ):
        self.data_files_df.to_hdf(
            path,
            key=data_files_key,
            mode="w",
        )
        self.annotations_df.to_hdf(
            path,
            key=annotations_key,
            mode="a",
        )

        if len(self.cloud_files_info) > 0:
            payload = [c.dict() for c in self.cloud_files_info]
            tmp_df = pd.DataFrame(payload)
            tmp_df.to_hdf(path, key=CLOUD_FILES_INFO_KEY, mode="a")

    def upload_data_files(
        self,
        client: HLClient,
        data_source_id: int,
        progress=False,
        data_file_dir: Union[str, Path] = "",
        multipart_filesize: Optional[str] = None,
    ):
        data_file_dir_path = Path(data_file_dir)
        append_data_file_dir = lambda f: str(data_file_dir_path / f)
        self.data_files_df.loc[:, "filename"] = self.data_files_df.filename.map(append_data_file_dir)

        data_file_path_to_id, failed_data_file_paths = create_data_files(
            client,
            self.data_files_df.filename.values,
            data_source_id,
            progress=progress,
            multipart_filesize=multipart_filesize,
        )

        old_data_file_ids = self.data_files_df.data_file_id.values
        self.data_files_df.loc[:, "data_file_id"] = self.data_files_df.filename.map(data_file_path_to_id)

        new_data_file_ids = self.data_files_df.data_file_id.values
        old_to_new_data_file_ids = {o: n for o, n in zip(old_data_file_ids, new_data_file_ids)}
        self.annotations_df.loc[:, "data_file_id"] = self.annotations_df.data_file_id.map(
            old_to_new_data_file_ids
        )

        LOG.debug(
            f"{len(data_file_path_to_id)} succeeded, {len(failed_data_file_paths)} failed -> {failed_data_file_paths}"
        )

    @classmethod
    def read_training_config(
        cls,
        hl_client: HLClient,
        training_config: TrainingConfigType,
        dataset_dir: Path,
        page_size=100,
    ) -> Dict[str, "Dataset"]:

        dataset_split_dict = {}
        for split in ("train", "dev", "test"):
            if training_config.data[f"datasets_{split}"]:
                dataset_ids = [d.id for d in training_config.data[f"datasets_{split}"]]
            else:
                continue

            _datasets = []

            for dataset_id in dataset_ids:
                cache_path = dataset_dir / f"records_{dataset_id}.json"
                if cache_path.exists():
                    ds = Dataset.read_from(
                        DatasetFormat.JSON,
                        path=cache_path,
                    )

                elif dataset_in_cloud(hl_client, dataset_id)[0]:
                    dataset_format = DatasetFormat.AWS_S3
                    ds = Dataset.read_from(
                        dataset_format,
                        dataset_id=dataset_id,
                        client=hl_client,
                    )
                    ds.write_json(cache_path)
                else:
                    dataset_format = DatasetFormat.HIGHLIGHTER_DATASET
                    ds = Dataset.read_from(
                        dataset_format,
                        dataset_id=dataset_id,
                        client=hl_client,
                        page_size=page_size,
                    )
                    ds.write_json(cache_path)

                _datasets.append(ds)
            if len(_datasets) > 1:
                dataset = Dataset.combine(_datasets)
            else:
                dataset = _datasets[0]

            dataset.data_files_df.loc[:, "split"] = split
            dataset_split_dict[split] = dataset

        assert "train" in dataset_split_dict
        assert any([s in dataset_split_dict for s in ("dev", "test")])
        return dataset_split_dict

    def interpolate_from_key_frames(
        self,
        frac: float = 1.0,
        frame_save_dir: Optional[Path] = None,
        source_video_dir: Optional[Path] = None,
    ):
        i_adf, i_ddf = interpolate_pixel_locations_between_frames(
            self.annotations_df,
            self.data_files_df,
            frame_save_dir=frame_save_dir,
            source_video_dir=source_video_dir,
        )
        return Dataset(annotations_df=i_adf, data_files_df=i_ddf)



# Declaration order is based on the dependency on static references
#
# To Do
# ~~~~~
# - None, yet !

from .utilities import (
    get_format, get_h264_decoder, get_h264_encoder, get_h264_encoder_options,
    GStreamerError, gst_initialise, enable_opencv, process_video
)

from .video_reader import VideoReader

from .video_camera_reader import VideoCameraReader

from .video_file_reader import VideoFileReader

from .video_file_writer import VideoFileWriter

from .video_stream_reader import VideoStreamReader

from .video_stream_writer import VideoStreamWriter

from .rtsp_io import VideoReadRTSP

from .scheme_rtsp import DataSchemeRTSP



#!/usr/bin/env python3
#
# Aiko Service: HyperSpace(Category(Dependency))
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# HyperSpace is a LifeCycleManager of Categories (category.py and lifecycle.py)
# Create, Read, Update and Destroy a network graph of Categories
# Add to, List and Remove from any Category, which is a group of Entries
# Entries can be either a Category or a Dependency
# Category paths support a tree-like structure "category_a/category_b/entry_c"
#
# Usage: Distributed Actor
# ~~~~~~~~~~~~~~~~~~~~~~~~
#   export HYPERSPACE_RANDOM_HASH=false    # use incrementing hash value
#
#   ./hyperspace.py run     [storage_url]  # file, in-memory, sqlite, mqtt
#   ./hyperspace.py dump
#   ./hyperspace.py exit
#
#   ./hyperspace.py add     ENTRY_PATH
#   ./hyperspace.py create  CATEGORY_PATH
#   ./hyperspace.py destroy CATEGORY_PATH
#   ./hyperspace.py link    ENTRY_PATH_NEW ENTRY_PATH_EXIST    # TODO: "link"
#   ./hyperspace.py list    [ENTRY_PATH] [--long_format | -l] [--recursive | -r]
#   ./hyperspace.py remove  ENTRY_PATH_EXIST                   # TODO: "unlink"
#   ./hyperspace.py update  ENTRY_PATH -p protocol -t key_0=value_0
#
# Usage: Bootstrap
# ~~~~~~~~~~~~~~~~
#   export HYPERSPACE_RANDOM_HASH=false  # use incrementing hash value
#
#   ./hyperspace.py initialize
#   ./hyperspace.py ln      new_link entry        # dependency | category
#   ./hyperspace.py ls      [-l] [-n] [-r] [path]
#   ./hyperspace.py mk      dependency            # dependency
#   ./hyperspace.py mkdir   category              # category only
#   ./hyperspace.py rm      entry                 # dependency | category
#   ./hyperspace.py storage [-s]
#
# Functions
# ~~~~~~~~~
#   _initialize()
#   _ln(link_path, target_path)
#   _ls(path, long_format, dependency_count, recursive))
#   _mk(name)
#   _mkdir(name)
#   _rm(name)
#   _storage(sort_by_name)
#
# To Do
# ~~~~~
# - Separate the design and implementation of HyperSpace and Storage
#   - Move HyperSpace file-system based implementation into StorageFile ?
#   - HyperSpace: create(): Category with different HyperSpace root ?
#   - Storage:    create(): Category or Dependency (Definition and Contents)
#
# - Implement "aiko_hyperspace repl" for interactive CRUD, etc
#
# * FIX: Consolidate service.py:ServiceFilter into ServiceFields (and use this)
#
# * FIX: "service.py:service_args()" to include "owner" --> "run_command()"
#
# * FIX: AttributeError: 'ServiceRemoteProxy' object has no attribute 'create'
#        do_command(HyperSpace, ServiceFilter(...),  # But HyperSpaceImpl works
#            lambda hyperspace: hyperspace.create(), ...)      # create() fails
#
# * HyperSpace as-a CategoryManager as-a LifeCycleManager as-a Category
#
# * HyperSpace/StorageFile uses FileSystemEventPatternMatch in "scheme_file.py"
#
# * aiko_hyperspace export output_filename.hyperspace --> hyperspace commands
#
# * Validate hyperspace structure
#   * Checks "storage/tracked_paths" is correct (using predictable hash)
#   * Checks file-system ensuring everything linked together properly
#   * Checks file-system ensuring that soft-links aren't "dead"
#   * Checks that file contents are correct (using predicatable data)
#   * Checks that directory contents are correct
#   * Checks file-system against a known good "backed-up" copy
#
# - Consider current "storage/tracked_paths" versus reference counting ?
#
# * ProcessManager bootstraps (read only) by directly using Python functions
#   * Implement HyperSpace(Actor) CRUD API ... used by ProcessManager bootstrap
#     * File-system listener --> event
#
# * Registrar(Actor) requires Category plus implement "(update)" for all fields
#
# * HyperSpace(Actor) requires Category and Dependencies
#   - Services that implement Category assign tag "category=true" (temporary)
#
# * Aiko Dashboard plug-in for HyperSpace(Actor), e.g REPL and tree view
#
# * Implement Category "owner" field and populate automatically
#   * Implement "destroy --force" flag to destroy other owner's Categories
#   * System Categories owned by "aiko"
#
# * Consider only one primary HyperSpace per host (refactor Registrar code)
#   * Don't allow two HyperSpace with the same name
# * HyperSpace primary / watchdog: monitor and relaunch primary
# * Unify HyperSpace on different hosts in the same namespace ?
#     * All the home/office servers plus handle mobile laptops (discovery) ?
#
# * HyperSpace uses Leases to enable lazy Category loading, running, unloading
#   - "aiko_hyperspace run" option for Category lease-time and maximum loaded
#   - ProcessManager(LifeCycleManager): Complete Service LifeCycle State
#     - Improvements to "src/aiko_services/main/state.py"
#   - ProcessManager(LifeCycleManager): Lazy loading of HyperSpace entries
#
# - StorageManager (content, definition) backed by ...
#   - ❌ In-memory, ✅ File-system, ❌ SQLite3, ❌ MQTT
# - HyperSpace (links) backed by ...
#   - ❌ In-memory, ✅ File-system, ❌ SQLite3, ❌ MQTT, ❌ Knowledge Graph

from abc import abstractmethod
import click
import errno
import hashlib
import os
from pathlib import Path
import stat
import sys

from aiko_services.main import *
from aiko_services.main.utilities import *

__all__ = ["HyperSpace"]

ACTOR_TYPE = "hyperspace"
VERSION = 0
PROTOCOL = f"{SERVICE_PROTOCOL_AIKO}/{ACTOR_TYPE}:{VERSION}"

_HASH_LENGTH = 12  # 6 bytes = 12 hex digits
_PATH_DELIMITER = "/"
_RESPONSE_TOPIC = f"{aiko.topic_in}"
_ROOT_FILENAME = ".root"
_STORAGE_FILENAME = "storage"
_TRACKED_PATHNAME = f"{_STORAGE_FILENAME}/tracked_paths"
_HASH_PATHNAME = f"{_STORAGE_FILENAME}/hash_counter"

# --------------------------------------------------------------------------- #

class HyperSpace(Category, Actor):
    Interface.default(
        "HyperSpace", "aiko_services.main.hyperspace.HyperSpaceImpl")

    @abstractmethod
    def create(self, category_path):
        pass

    @abstractmethod
    def destroy(self, category_path):
        pass

    @abstractmethod
    def dump(self):
        pass

# --------------------------------------------------------------------------- #

class HyperSpaceImpl(HyperSpace):
    def __init__(self, context, storage_url):
        context.call_init(self, "Actor", context)
        context.call_init(self, "Category", context)

        self.category = context.get_implementation("Category")        # methods

        self.share.update({                             # Inherit from Category
            "source_file": f"v{VERSION}⇒ {__file__}",
            "storage_url": storage_url,
            "metrics": {
                "created": 0,
                "running": 1 + len(self.entries)   # Including self 😅
            #   "time_started": 0            # TODO: UTC time started --> Actor
            },                               #       or time.monotonic()
        })
        self.ec_producer = ECProducer(self, self.share)
        self.ec_producer.add_handler(self._ec_producer_change_handler)

    #   self.thread = Thread(target=self._run, daemon=True, name=_THREAD_NAME)
    #   self.thread.start()                           # TODO: Use ThreadManager

    #   if not storage_url:
    #       storage_url = PATHNAME_OF_CURRENT_WORKING_DIRECTORY
    #   self._load_hyperspace(storage_url)

    def _category_path_iterator(self, category_path, delimiter=_PATH_DELIMITER):
        for name in category_path.split(delimiter):
            yield name

    def _category_traverse(self, category_path, delimiter=_PATH_DELIMITER):
        dirname, basename = dir_base_name(category_path)
        current_category = self
        if dirname == ".":
            return current_category, basename

        for category_name in self._category_path_iterator(dirname):
            if category_name in current_category.entries:
                entry = current_category.entries[category_name]
                if entry.service_filter.protocol != CATEGORY_PROTOCOL:
                    return None, None
                current_category = entry
            else:
                return None, None
        return current_category, basename

    def add(self, entry_path,
        service_filter, lifecycle_manager_url=None, storage_url=None):

        category, entry_name = self._category_traverse(entry_path)
        if category:
            if service_filter and not isinstance(service_filter, ServiceFilter):
                service_filter = ServiceFilter(*service_filter)
            if service_filter.name == "*":
                service_filter.name = entry_name
            self.category.add(category, entry_name,
                service_filter, lifecycle_manager_url, storage_url)

    def create(self, category_path):
        current_category = self
        category_created = False
        tags = ["ec=true"]

        for category_name in self._category_path_iterator(category_path):
            if category_name in current_category.entries:
                entry = current_category.entries[category_name]
                if entry.service_filter.protocol != CATEGORY_PROTOCOL:
                    break
                category = entry
            else:
                init_args = actor_args(
                    category_name, None, None, CATEGORY_PROTOCOL, tags)
                init_args["service_filter"] = ServiceFilter(
                    name=category_name, protocol=CATEGORY_PROTOCOL, tags=[])
                category = compose_instance(CategoryImpl, init_args)
                current_category.entries[category_name] = category
                category_created = True
            current_category = category

        if category_created:
            self.ec_producer.update("entries", len(self.entries))

# TODO: Categories destroy() versus Dependencies remove() ?  Check protocol ?
# TODO: Recursively remove Categories ... using "limit" argument
# TODO: Replace "destroy()" with "remove()", only destroy when last reference
# TODO: Remove Dependencies and clean-up any resources ?

    def destroy(self, category_path):
        category_parent, category_name = self._category_traverse(category_path)
        self._destroy(category_parent, category_name)

    def _destroy(self, category_parent, category_name):
        if category_parent and category_name in category_parent.entries:
            category = category_parent.entries[category_name]
            if category.service_filter.protocol == CATEGORY_PROTOCOL:
                del category_parent.entries[category_name]
                aiko.process.remove_service(category.service_id)
                self.ec_producer.update("entries", len(self.entries))

    def dump(self):
        if len(self.entries):
            state = f" ...\n{self}"
        else:
            state = ": no entries"
        self.logger.info(f"Dump state{state}")

    def _ec_producer_change_handler(self, command, item_name, item_value):
        if item_name == "log_level":
            self.logger.setLevel(str(item_value).upper())

    def exit(self):
        aiko.process.terminate()

    def is_type(self, type_name):
        if type_name.lower() == "hyperspace":
            return True
        return self.category.is_type(self, type_name)

    def list(self, topic_path_response, entry_path=None, recursive=False):
        if not isinstance(recursive, bool):
            recursive = recursive.lower() in ("true", "t")

        category, entry_name = self._category_traverse(entry_path)
        if category and entry_name in category.entries:
            entry = category.entries[entry_name]
            if entry.service_filter.protocol == CATEGORY_PROTOCOL:
                category = entry

        def get_entry_records(entry_records, category, category_name, level):
            if level > 0:
                entry_records += [f"{-level} {category_name}"]
            entry_records += category._get_entry_records(None, level)
            if recursive:
                for entry_name, entry in category.entries.items():
                    if entry.is_type("Category"):
                        get_entry_records(
                            entry_records, entry, entry_name, level+1)

        entry_records = []
        get_entry_records(entry_records, category, None, 0)
        category._list_publish(topic_path_response, entry_records)

# TODO: Categories destroy() versus Dependencies remove() ?  Check protocol ?

    def remove(self, entry_path):
        category_parent, entry_name = self._category_traverse(entry_path)
        if category_parent and entry_name in category_parent.entries:
            entry = category_parent.entries[entry_name]
            if entry.service_filter.protocol == CATEGORY_PROTOCOL:
                self._destroy(category_parent, entry_name)
            else:
                self.category.remove(category_parent, entry_name)

    def __str__(self):
        result = ""
        for entry_name, entry in self.entries.items():
            entry_type = "/" if entry.is_type("Category") else ""
            new_line = "\n" if result else ""
            result += f"{new_line}  {entry_name}{entry_type}"
        return result

    def update(self, entry_path, service=None,
        service_filter=None, lifecycle_manager_url=None, storage_url=None):

        category_parent, entry_name = self._category_traverse(entry_path)
        if category_parent and entry_name in category_parent.entries:
            self.category.update(category_parent, entry_name, service,
                service_filter, lifecycle_manager_url, storage_url)

# --------------------------------------------------------------------------- #
# HyperSpace Storage: File-system based persistance

_hash_counter = 0

# Initialize storage directory and .root symlink

def _check_root_symlink():
    root_pathname = Path.cwd() / _ROOT_FILENAME
    if not os.path.lexists(root_pathname):
        raise FileExistsError(
            f'HyperSpace: Symlink "./{_ROOT_FILENAME}" must exist')
    else:
        if not os.path.islink(root_pathname):
            raise FileExistsError(
                f'HyperSpace: File "./{_ROOT_FILENAME}" must be a symlink')

def _initialize():
    cwd = Path.cwd()
    root_pathname = cwd / _ROOT_FILENAME
    if not root_pathname.is_symlink():
        root_pathname.symlink_to(cwd)
        print(f"Created {_ROOT_FILENAME} --> {cwd}")

    storage_pathname = cwd / _STORAGE_FILENAME
    if not storage_pathname.is_dir():
        storage_pathname.mkdir(parents=True)
        print(f"Created {_STORAGE_FILENAME}")

    global _hash_counter
    hash_pathname = cwd / _HASH_PATHNAME
    if not hash_pathname.exists():
        hash_pathname.write_text("0")
    try:
        _hash_counter = int(hash_pathname.read_text())
    except ValueError:
        _hash_counter = 0

# Generate hash UID, persisting the hash_counter when using incremental mode

def _generate_hash():
    global _hash_counter
    if os.environ.get("HYPERSPACE_RANDOM_HASH", "true").lower() == "true":
        hash = hashlib.sha256(os.urandom(16)).hexdigest()
    else:
        hash = str(_hash_counter).zfill(_HASH_LENGTH)
        _hash_counter += 1
        Path(_HASH_PATHNAME).write_text(f"{_hash_counter}\n")
    return hash[- _HASH_LENGTH:]

# Clean up empty storage directories up to root

def _clean_storage(tracked_directory):
    tracked_path = Path(tracked_directory)
    storage_path = Path(_STORAGE_FILENAME)
    while tracked_path != storage_path and tracked_path != Path("."):
        if tracked_path.is_dir() and not any(tracked_path.iterdir()):
            tracked_path.rmdir()
            tracked_path = tracked_path.parent
        else:
            break

# Generate a unique storage path avoiding collisions

def _create_path():
    storage_path = Path(_STORAGE_FILENAME)
    existing_paths = set()
    if Path(_TRACKED_PATHNAME).exists():
        existing_paths.update(Path(_TRACKED_PATHNAME).read_text().splitlines())
    while True:
        hash = _generate_hash()
        path_parts = [hash[i:i+2] for i in range(0, len(hash), 2)]
        new_path = storage_path.joinpath(*path_parts)
        if not new_path.exists() and str(new_path) not in existing_paths:
            return str(new_path)

# Normalize path to relative from CWD

def _normalize_path(target_path):
    absolute_target_path = Path(target_path).resolve()
    cwd = Path.cwd()
    try:
        relative_path = absolute_target_path.relative_to(cwd)
        return f"./{relative_path}"
    except ValueError:
        return str(absolute_target_path)

# Compute relative path using os.path.relpath

def _relative_path(target_path, start_path):
    return os.path.relpath(target_path, start=start_path)

# Track a created storage path

def _track_path(path):
    relative_path = _normalize_path(path)
    Path(_TRACKED_PATHNAME).parent.mkdir(parents=True, exist_ok=True)
    with open(_TRACKED_PATHNAME, "a+") as tracked_paths_file:
        tracked_paths_file.seek(0)
        tracked_paths = tracked_paths_file.read().splitlines()
        if relative_path not in tracked_paths:
            tracked_paths_file.write(relative_path + "\n")

# Untrack a removed storage path

def _untrack_path(path):
    relative_path = _normalize_path(path)
    tracked_path = Path(_TRACKED_PATHNAME)
    if tracked_path.is_file():
        tracked_paths = tracked_path.read_text().splitlines()
        tracked_path.write_text(
            "\n".join(l for l in tracked_paths if l != relative_path))

# Create new link to an existing dependency or category
#
# ./hyperspace.py ln new_link entry  # dependency | category

def _ln(link_path, target_path):
    if not Path(target_path).exists():
        print(f'Error: target "{target_path}" does not exist', file=sys.stderr)
        return 1
    resolved_path = Path(target_path).resolve()
    cwd = Path.cwd()
    if _STORAGE_FILENAME not in resolved_path.parts:
        print(f'Error: target "{target_path}" is not in the storage directory',
            file=sys.stderr)
        return 1
    base_directory = Path(link_path).parent or Path(".")
    base_root = base_directory / _ROOT_FILENAME
    link_base = base_root if base_root.is_symlink() else Path(_ROOT_FILENAME)

    # Determine path under ".root/storage/"
    try:
        relative_path = resolved_path.relative_to(cwd / _STORAGE_FILENAME)
        relative_storage_path =  \
            f"{_STORAGE_FILENAME}/{relative_path.as_posix()}"
    except ValueError:
        relative_storage_path = resolved_path.relative_to(cwd).as_posix()

    # Construct full path under .root
    dot_root_path = link_base / relative_storage_path
    relative_path = _relative_path(str(dot_root_path), str(base_directory))
    try:
        Path(link_path).symlink_to(relative_path)
    except FileExistsError:
        print(f'Error: "{link_path}" already exists', file=sys.stderr)
        return 1

# List categories and depedencies in a tree-style recursive format
#
# ./hyperspace.py ls [-l] [-n] [-r] [path]

def _ls(path, long_format=False, entry_count=False, recursive=False):
    def _get_hash_path(link):
        absolute_target_path = Path(link).resolve()
        try:
            relative_path =  \
                absolute_target_path.relative_to(Path.cwd() / _STORAGE_FILENAME)
            return str(relative_path)
        except ValueError:
            return ""

    def _file_count(link_directory):
        target_path = Path(link_directory).resolve()
        return sum(1 for f in target_path.rglob("*")  \
                      if f.is_file() and f.stat().st_size > 0)

    def _list_links(current, indent=""):
        for directory_entry in sorted(Path(current).iterdir()):
            if not directory_entry.is_symlink():
                continue
            entry_name = directory_entry.name
            if entry_name.startswith("."):  # Skip hidden directory entries
                continue
            hash = _get_hash_path(directory_entry)
            if directory_entry.is_dir():
                if long_format:
                    print(f"{hash}  {indent}{entry_name}/", end="")
                else:
                    print(f"{indent}{entry_name}/", end="")
                if entry_count:
                    count = _file_count(directory_entry)
                    if count > 0:
                        print(f" ({count})", end="")
                print()
                if recursive and not entry_name.startswith("."):
                    _list_links(directory_entry, indent + "  ")
            else:
                if long_format:
                    print(f"{hash}  {indent}{entry_name}")
                else:
                    print(f"{indent}{entry_name}")

    _list_links(path)

# Create a Dependency (file)
#
# ./hyperspace.py mk entry  # Dependency only

def _mk(name):
    path = Path(_create_path())
    path.parent.mkdir(parents=True, exist_ok=True)
    path.touch()
    try:
        Path(name).symlink_to(f"{_ROOT_FILENAME}/{path}")
    except FileExistsError:
        print(f'Error: "{name}" already exists', file=sys.stderr)
        return 1
    _track_path(path)

# Create a category (directory)
#
# ./hyperspace.py mkdir category  # category only

def _mkdir(name):
    path = Path(_create_path())
    path.mkdir(parents=True, exist_ok=True)
    rel_back = _relative_path(str(Path.cwd() / _ROOT_FILENAME), path)
    try:
        Path(path, _ROOT_FILENAME).symlink_to(rel_back)
    except FileExistsError:
        print(f'Error: internal .root link in "{path}" already exists',
            file=sys.stderr)
        return 1
    try:
        Path(name).symlink_to(f"{_ROOT_FILENAME}/{path}")
    except FileExistsError:
        print(f'Error: "{name}" already exists', file=sys.stderr)
        return 1
    _track_path(path)

# Delete a Dependency (file) or Category (directory)
#
# ./hyperspace.py rm entry  # Dependency | Category

def _rm(name):
    path = Path(name)
    if path.is_symlink():
        absolute_target_path = path.resolve()
        path.unlink()
        cwd = Path.cwd()
        try:
            storage_path = cwd / _STORAGE_FILENAME
            relative_path = absolute_target_path.relative_to(storage_path)
        except ValueError:
            return

        others = [x for x in cwd.rglob("*")  \
                     if x.is_symlink() and x.resolve() == absolute_target_path]
        if not others:
            _untrack_path(str(absolute_target_path))
            if absolute_target_path.is_dir():
                for sub in absolute_target_path.rglob("*"):
                    sub.unlink() if sub.is_file() else sub.rmdir()
                absolute_target_path.rmdir()
            else:
                absolute_target_path.unlink()
            _clean_storage(absolute_target_path.parent)
    else:
        print(f'Error: "{name}" is not a symbolic link', file=sys.stderr)
        return 1

# List storage in reverse-mapped form with deduplication and optional name sort
#
# ./hyperspace.py storage [-s]

def _storage(sort_by_name=False):
    tracked_paths = set()
    cwd = Path.cwd()
    for path in cwd.rglob("*"):
        if path.is_symlink():
            absolute_target_path = path.resolve()
            try:
                absolute_path = cwd / _STORAGE_FILENAME
                relative_path = absolute_target_path.relative_to(absolute_path)
                tracked_paths.add((str(relative_path), path.name))
            except ValueError:
                continue
    items = sorted(tracked_paths, key=lambda x: x[1] if sort_by_name else x[0])
    for relative_path, name in items:
        print(f"{relative_path}  {name}")

# --------------------------------------------------------------------------- #
# HyperSpace CLI: Bootstrap commands

@click.group()

def main():
    """Create, Read/List, Update and Destroy HyperSpace"""

    subcommand = click.get_current_context().invoked_subcommand
    subcommands_which_run_anywhere = ["create", "dump", "exit", "initialize"]

    if subcommand not in subcommands_which_run_anywhere:
        diagnostic = None
        try:
            _check_root_symlink()
        except FileExistsError as file_exists_error:
            diagnostic = str(file_exists_error)
        if diagnostic:
            advice = 'Consider running "aiko_hyperspace initialize"'
            raise SystemExit(f"{diagnostic}\n{advice}")

# ./hyperspace.py initialize

@main.command(name="initialize")

def initialize_command():
    "Initialize HyperSpace: .root and storage/"
    _initialize()

# ./hyperspace.py ln new_link_entry existing_entry

@main.command(name="ln")
@click.argument("link_path")
@click.argument("target_path")

def ln_command(link_path, target_path):
    "Create new link to an existing dependency or category"
    _ln(link_path, target_path)

# ./hyperspace.py ls [-l] [-n] [-r] [path]

@main.command(name="ls")
@click.argument("path", required=False, default=".")
@click.option("--entry_count", "-c", default=False, is_flag=True,
    help="Show category's entry count")
@click.option("--long_format", "-l", default=False, is_flag=True,
    help="Long format with hash identifiers")
@click.option("--recursive", "-r", default=False, is_flag=True,
    help="Recursive listing")

def ls_command(path, entry_count, long_format, recursive):
    "List dependencies (files) and categories (directories)"
    _ls(path, long_format, entry_count, recursive)

# ./hyperspace.py mk entry  # dependency only

@main.command(name="mk")
@click.argument("name")

def mk_command(name):
    "Create a dependency (file)"
    _mk(name)

# ./hyperspace.py mkdir category  # category only

@main.command(name="mkdir")
@click.argument("name")

def mkdir_command(name):
    "Create a category (directory)"
    _mkdir(name)

# ./hyperspace.py rm entry  # dependency | category

@main.command(name="rm")
@click.argument("name")

def rm_command(name):
    "Remove a dependency or category"
    _rm(name)

# ./hyperspace.py storage [-s]

@main.command(name="storage")
@click.option("--sort_by_name", "-s", default=False, is_flag=True,
    help="Sort by name")

def storage_command(sort_by_name):
    "List entry storage/tracked_paths and linked entries"
    _storage(sort_by_name)

# --------------------------------------------------------------------------- #
# HyperSpace CLI: Distributed Actor commands
#
# To Do
# ~~~~~
# add
# ~~~
# * Protocol well-known short names, e.g "category" --> CATEGORY_PROTOCOL
# * Created and modified time
# - Owner defaults to current user, ACLs "u:name g:name o:name rwx:name ..." ?
# - ServiceFilter fields (see aiko_hyperspace create)
#
# create
# ~~~~~~
# * Differentiate Categories with same name, e.g ENTRY_PATH tag ?
# * Protocol well-known short names, e.g "category" --> CATEGORY_PROTOCOL
# * Created and modified time
# - Owner defaults to current user, ACLs "u:name g:name o:name rwx:name ..." ?
# - ServiceFilter fields ...
#   -n,   --service_name TEXT           # Service name
#   -p,   --protocol TEXT               # CATEGORY_PROTOCOL (default)
#   -tr,  --transport TEXT              # Service transport (MQTT default)
# * -o,   --owner TEXT                  # Service owner     (current user)
#   -t,   --tags TEXT                   # Service tags
# * -lcm, --lifecycle_manager_url TEXT  # LifeCycleManager URL
# * -s,   --storage_url TEXT            # Storage URL (file, sqlite, MQTT, etc)
#
# destroy
# ~~~~~~~
# * Clean-up all Process, Service, Actor, Dependency, Category resources
# * Only destroy() Categories with a LifeCycleManager of "HyperSpaceImpl.self"
# - Regex, e.g wildcards "*", "+", ...
# - [r]ecursive [level]  # default: 1, * means no limit
# - Only owner can destroy ... implement ACLs later
#
# list
# ~~~~
# * Specific "entry_path"
# * Regex, e.g wildcards "*", "+", ...
# * [c]ount  # Category Entry count
# * [l]ong_format
# * [r]ecursive [level]  # default: 1 or * ? (* means no limit)
#
# remove
# ~~~~~~
# - Regex, e.g wildcards "*", "+", ...
# - [r]ecursive [level]  # default: 1, * means no limit
# - Only owner can remove ... ACLs later
#
# update
# ~~~~~~
# * category_path tree: category/category/category
# * Protocol well-known short names, e.g "category" --> CATEGORY_PROTOCOL
# - Regex, e.g wildcards "*", "+", ...
# - ServiceFilter fields (see aiko_hyperspace create)
# - Only owner can update ... implement ACLs later

@main.command(name="add", no_args_is_help=True)
@click.argument("entry_path", type=str, required=True, default=None)
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")
@click.option("--service_name", "-n", type=str, default="*",
    help="Service name")
@click.option("--protocol", "-p", type=str, default="*",
    help="Service protocol")
@click.option("--transport", "-tr", type=str, default="*",
    help="Service transport")
@click.option("--owner", "-o", type=str, default="*",
    help="Service owner")
@click.option("--tags", "-t", type=str, multiple=True, default=None,
    help="Service tags")
@click.option("--lifecycle_manager_url", "-lcm", type=str, default=None,
    help="LifeCycleManager URL")
@click.option("--storage_url", "-s", type=str, default=None,
    help="Storage URL")

def add_command(entry_path, hyperspace_name,
    service_name, protocol, transport, owner, tags,
    lifecycle_manager_url, storage_url):

    """Add HyperSpace Entry

    aiko_hyperspace add ENTRY_PATH

    \b
    • ENTRY_PATH: Entry PATH
    """

    tags = tags if tags else []                 # Assign default tags value
    service_filter = ServiceFilter(             # TODO: Or use ServiceFields ??
        "*", service_name, protocol, transport, owner, tags)

    do_command(Category,
        ServiceFilter(name=hyperspace_name, protocol=PROTOCOL),
        lambda hyperspace: hyperspace.add(entry_path,
            service_filter, lifecycle_manager_url, storage_url), terminate=True)
    aiko.process.run()

@main.command(name="create", no_args_is_help=True)
@click.argument("category_path", type=str, required=True, default=None)
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")

def create_command(category_path, hyperspace_name):
    """Create HyperSpace Category

    aiko_hyperspace create [-hn HYPERSPACE_NAME] CATEGORY_PATH

    \b
    • CATEGORY_PATH: Category path
    """

    do_command(HyperSpaceImpl,
        ServiceFilter(name=hyperspace_name, protocol=PROTOCOL),
        lambda hyperspace: hyperspace.create(category_path), terminate=True)
    aiko.process.run()

@main.command(name="destroy", no_args_is_help=True)
@click.argument("category_path", type=str, required=True, default=None)
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")

def destroy_command(category_path, hyperspace_name):
    """Destroy HyperSpace Category

    aiko_hyperspace destroy [-hn HYPERSPACE_NAME] CATEGORY_PATH

    \b
    • CATEGORY_PATH: Category path
    """

    do_command(HyperSpaceImpl,
        ServiceFilter(name=hyperspace_name, protocol=PROTOCOL),
        lambda hyperspace: hyperspace.destroy(category_path), terminate=True)
    aiko.process.run()

@main.command(name="dump", help="Dump HyperSpace state")
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")

def dump_command(hyperspace_name):
    do_command(HyperSpace,
        ServiceFilter(name=hyperspace_name, protocol=PROTOCOL),
        lambda hyperspace: hyperspace.dump(), terminate=True)
    aiko.process.run()

@main.command(name="exit", help="Exit HyperSpace")
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")

def exit_command(hyperspace_name):
    do_command(HyperSpace,
        ServiceFilter(name=hyperspace_name, protocol=PROTOCOL),
        lambda hyperspace: hyperspace.exit(), terminate=True)
    aiko.process.run()

@main.command(name="list")
@click.argument("entry_path", type=str, required=False, default=None)
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")
@click.option("--long_format", "-l", is_flag=True,
    help="Long format with Service, LifeCycleManager URL, Storage URL")
@click.option("--recursive", "-r", is_flag=True,
    help="Recursively list Category entries")

def list_command(entry_path, hyperspace_name, long_format, recursive):
    """List HyperSpace Entries

    aiko_category list ENTRY_PATH [-l] [-r]

    \b
    • ENTRY_PATH: Entry path
    """

    CategoryImpl.list_command(
        hyperspace_name, entry_path, long_format, recursive, PROTOCOL)
    aiko.process.run()

@main.command(name="remove", no_args_is_help=True)
@click.argument("entry_path", type=str, required=True, default=None)
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")

def remove_command(entry_path, hyperspace_name):
    """Remove HyperSpace Entry

    aiko_hyperspace remove [-hn HYPERSPACE_NAME] ENTRY_PATH

    \b
    • ENTRY_PATH: Entry path
    """

    do_command(HyperSpaceImpl,
        ServiceFilter(name=hyperspace_name, protocol=PROTOCOL),
        lambda hyperspace: hyperspace.remove(entry_path), terminate=True)
    aiko.process.run()

@main.command(name="run")
@click.argument("storage_url", required=False, default=None)  # TODO: Implement
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")
@click.option("--protocol", "-p", type=str, default=PROTOCOL,
    help="Service protocol")
@click.option("--transport", "-tr", type=str, default=None,
    help="Service transport")
# @click.option("--owner", "-o", type=str, default=None,      # TODO: Implement
#     help="Service owner")
@click.option("--tags", "-t", type=str, multiple=True, default=None,
    help="Service tags")

def run_command(storage_url, hyperspace_name, protocol, transport, tags):
    """Run HyperSpace

    aiko_hyperspace run [-hn HYPERSPACE_NAME] [STORAGE_URL]

    \b
    • STORAGE_URL: HyperSpace storage location, e.g file://...
    """

    hyperspace_name = hyperspace_name if hyperspace_name else get_hostname()

    tags = list(tags) + ["ec=true"]
    init_args = actor_args(
        hyperspace_name, None, None, protocol, tags, transport)
    init_args["storage_url"] = storage_url
    hyperspace = compose_instance(HyperSpaceImpl, init_args)
    aiko.process.run()

@main.command(name="update", no_args_is_help=True)
@click.argument("entry_path", type=str, required=True, default=None)
@click.option("--hyperspace_name", "-hn", type=str, default=None,
    help="HyperSpace name, default is the local hostname")
@click.option("--service_name", "-n", type=str, default="0:",     # None --> 0:
    help="Service name")
@click.option("--protocol", "-p", type=str, default="0:",         # None --> 0:
    help="Service protocol")
@click.option("--transport", "-tr", type=str, default="0:",       # None --> 0:
    help="Service transport")
@click.option("--owner", "-o", type=str, default="0:",            # None --> 0:
    help="Service owner")
@click.option("--tags", "-t", type=str, multiple=True, default=None,
    help="Service tags")
@click.option("--lifecycle_manager_url", "-lcm", type=str, default=None,
    help="LifeCycleManager URL")
@click.option("--storage_url", "-s", type=str, default=None,
    help="Storage URL")

def update_command(entry_path, hyperspace_name,
    service_name, protocol, transport, owner, tags,
    lifecycle_manager_url, storage_url):

    """Update HyperSpace Entry

    aiko_hyperspace update [-hn HYPERSPACE_NAME] ENTRY_PATH

    \b
    • ENTRY_PATH: Entry path
    """

    tags = tags if tags else []                 # Assign default tags value
    service_filter = ServiceFilter(             # TODO: Or use ServiceFields ??
        "*", service_name, protocol, transport, owner, tags)

    do_command(HyperSpaceImpl,
        ServiceFilter(name=hyperspace_name, protocol=PROTOCOL),
        lambda hyperspace: hyperspace.update(entry_path, None,
            service_filter, lifecycle_manager_url, storage_url), terminate=True)
    aiko.process.run()

if __name__ == "__main__":
    main()

# --------------------------------------------------------------------------- #



import logging
import uuid
from typing import List, Optional
from uuid import UUID

from gql.transport.exceptions import TransportQueryError
from sqlalchemy import Column, String
from sqlmodel import Field, Relationship, SQLModel

from highlighter.client import HLClient
from highlighter.client.base_models.base_models import DataSourceType
from highlighter.core.data_models.account_mixin import AccountMixin
from highlighter.core.gql_base_model import GQLBaseModel

logger = logging.getLogger(__name__)


class DataSource(SQLModel, AccountMixin, GQLBaseModel, table=True):
    name: str = Field(sa_column=Column(String))
    uuid: UUID = Field(default_factory=lambda: uuid.uuid4(), primary_key=True)
    data_files: list["DataFile"] = Relationship(back_populates="data_source")

    def save_to_cloud(self):
        cloud_data_source_results = None

        try:
            cloud_data_source_results = HLClient.get_client().data_source(
                return_type=DataSourceType,
                uuid=self.data_source_uuid,
            )
        except TransportQueryError as e:
            self.logger.info(
                f"data_source_id not found in Highlighter cloud: {self.data_source_id}, saving local data source to cloud"
            )

        if cloud_data_source_results is None:

            class CreateDataSourcePayload(GQLBaseModel):
                data_source: Optional[DataSourceType]
                errors: List[str]

            response = HLClient.get_client().create_data_source(
                return_type=CreateDataSourcePayload,
                uuid=str(self.uuid),
                name=self.name,
            )

            logger.info(f"DataSource saved to cloud with uuid {self.uuid} with response {response}")

            return response
        elif cloud_data_source_results.uuid != self.uuid:
            self.logger.info(
                f"data_source_id not found in Highlighter cloud: {self.data_source_id}, saving local data source to cloud"
            )
            with self.database.get_session() as session:
                self.uuid = cloud_data_source_results.uuid
                session.add(self)
                session.commit()
                session.refresh(self)



from datetime import datetime
from typing import List, Tuple
from uuid import UUID, uuid4

from shapely import geometry as geom

from highlighter import (
    OBJECT_CLASS_ATTRIBUTE_UUID,
    Annotation,
    DatumSource,
    Entity,
    Observation,
)
from highlighter.agent.capabilities import Capability, StreamEvent
from highlighter.core.data_models import DataSample


class MockDetector(Capability):
    """Mock detector that creates dummy detections on image or text source data"""

    class StreamParameters(Capability.StreamParameters):
        output_object_class_ids: List[UUID]

    def process_frame(self, stream, data_samples: List[DataSample]) -> Tuple[StreamEvent, dict]:
        parameters = self.stream_parameters(stream.stream_id)
        if len(data_samples) == 0:
            return StreamEvent.OKAY, {"entities": {}}

        entities = {}
        for data_sample in data_samples:
            if data_sample.content_type.startswith("image"):
                locations = [
                    geom.Polygon([(0, 0), (0, 10), (10, 10), (10, 0)]),
                    geom.Polygon([(100, 100), (100, 110), (110, 110), (110, 100)]),
                ]
            elif data_sample.content_type.startswith("text"):
                locations = [geom.LineString([(0, 0), (10, 0)]), geom.LineString([(100, 100), (110, 100)])]
            else:
                raise ValueError(
                    "MockDetector.process_frame() must be given either 'image' or 'text' data samples"
                )
            for location in locations:
                for object_class_id in parameters.output_object_class_ids:
                    entity = Entity()
                    entities[entity.id] = entity

                    annotation_datum_source = DatumSource(
                        confidence=1, frame_id=data_sample.media_frame_index
                    )
                    annotation = Annotation(
                        entity=entity,
                        location=location,
                        data_file_id=data_sample.data_file_id,
                        datum_source=annotation_datum_source,
                    )

                    observation_datum_source = DatumSource(
                        confidence=1, frame_id=data_sample.media_frame_index
                    )
                    Observation(
                        annotation=annotation,
                        attribute_id=OBJECT_CLASS_ATTRIBUTE_UUID,
                        value=object_class_id,
                        occurred_at=datetime.now(),
                        datum_source=observation_datum_source,
                    )
        return StreamEvent.OKAY, {"entities": entities}



import os
import sys
from logging.config import fileConfig

from alembic import context
from sqlalchemy import engine_from_config, pool
from sqlmodel import SQLModel

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from highlighter.core.config import HighlighterRuntimeConfig
from highlighter.core.data_models import *

# The alembic/env.py is not imported as part of normal
# Highlighter SDK operations. So this global instantiation of
# the HighlighterRuntimeConfig should only be executed when doing
# database migrations.
# https://alembic.sqlalchemy.org/en/latest/tutorial.html
hl_cfg = HighlighterRuntimeConfig.load()

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

config.set_main_option("sqlalchemy.url", "sqlite:///" + str(hl_cfg.agent.db_file()))

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
target_metadata = SQLModel.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def include_object(object, name, type_, reflected, compare_to):
    if type_ == "column":
        # Ignore default_factory or any callable default
        if hasattr(object, "default") and object.default is not None and callable(object.default.arg):
            object.default = None
    return True


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    connectable = engine_from_config(
        config.get_section(config.config_ini_section, {}),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            include_object=include_object,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



#!/usr/bin/env python3
#
# Usage
# ~~~~~
# AIKO_LOG_LEVEL=DEBUG ./storage.py start &
#
# AIKO_LOG_LEVEL=DEBUG ./storage.py test_command [argument_0]
#   Storage command: test_command(argument_0)
#
# AIKO_LOG_LEVEL=DEBUG ./storage.py test_request [request_0]
#   Storage response: [("request_0", [])]
#
# To Do
# ~~~~~
# - Rename "StorageImpl" to "StorageFile"
#
# - Consider GraphQL over MQTT !

from abc import abstractmethod
import click
import sqlite3

from aiko_services.main import *

__all__ = ["Storage"]

_VERSION = 0

ACTOR_TYPE = "storage"
PROTOCOL = f"{SERVICE_PROTOCOL_AIKO}/{ACTOR_TYPE}:{_VERSION}"

_LOGGER = aiko.logger(__name__)
_RESPONSE_TOPIC = f"{aiko.topic_in}"

# --------------------------------------------------------------------------- #

class Storage(Actor):
    Interface.default("Storage", "aiko_services.main.storage.StorageImpl")

    @abstractmethod
    def test_command(self, parameter):
        pass

    @abstractmethod
    def test_request(self, topic_path_response, request):
        pass

class StorageImpl(Storage):
    def __init__(self, context, database_pathname):
        context.call_init(self, "Actor", context)
        self.connection = sqlite3.connect(database_pathname)
        self.share["database_pathname"] = database_pathname
        self.share["source_file"] = f"v{_VERSION}⇒ {__file__}"

    def test_command(self, parameter):
        print(f"Storage command: test_command({parameter})")

    def test_request(self, topic_path_response, request):
        aiko.message.publish(topic_path_response, "(item_count 1)")
        aiko.message.publish(topic_path_response, f"({request})")

# --------------------------------------------------------------------------- #

@click.group()

def main():
    pass

@main.command(help="Start Storage")
@click.argument("database_pathname", default="aiko_storage.db")

def start(database_pathname):
    tags = ["ec=true"]
    init_args = actor_args(ACTOR_TYPE, protocol=PROTOCOL, tags=tags)
    init_args["database_pathname"] = database_pathname
    storage = compose_instance(StorageImpl, init_args)
    storage.run()

@main.command(name="test_command", help="Test Storage invoke command")
@click.argument("argument", default="argument_0", required=False)

def test_command(argument):
    do_command(Storage, ServiceFilter("*", "*", PROTOCOL, "*", "*", "*"),
        lambda storage: storage.test_command(argument), terminate=True)
    aiko.process.run()

@main.command(name="test_request", help="Test Storage invoke request")
@click.argument("request", default="request_0", required=False)

def test_request(request):
    def response_handler(response):
        print(f"Storage response: {response}")

    do_request(Storage, ServiceFilter("*", "*", PROTOCOL, "*", "*", "*"),
        lambda storage: storage.test_request(_RESPONSE_TOPIC, request),
        response_handler, _RESPONSE_TOPIC, terminate=True)
    aiko.process.run()

if __name__ == "__main__":
    main()

# --------------------------------------------------------------------------- #



import hashlib
import re


def sha512_of_content(content):
    sha512 = hashlib.sha512()
    sha512.update(content)
    return sha512.hexdigest()


def get_slug(name) -> str:
    return re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()


def stringify_if_not_null(attribute):
    return str(attribute) if attribute is not None else None



import warnings
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union
from uuid import UUID

import numpy as np
from PIL import Image
from shapely.geometry import Polygon

from ...core import GQLBaseModel, draw_annotations_on_image, resize_image

__all__ = ["DataFile"]

warnings.simplefilter("always", DeprecationWarning)  # Force showing deprecation warnings


class DataFile(GQLBaseModel):
    file_id: Optional[UUID] = None
    content_type: str
    content: Any
    recorded_at: datetime = datetime.now()
    media_frame_index: int = 0
    original_source_url: Optional[str] = None

    @classmethod
    def from_image(
        cls,
        content: Union[np.ndarray, Image.Image, str, Path],
        file_id: UUID,
        media_frame_index: int = 0,
        original_source_url: Optional[str] = None,
    ):
        if isinstance(content, (str, Path)):
            content = Image.open(content)

        assert isinstance(content, (np.ndarray, Image.Image))

        return cls(
            file_id=file_id,
            content_type="image",
            content=content,
            media_frame_index=media_frame_index,
            original_source_url=original_source_url,
        )

    def get_id(self) -> UUID:
        return self.file_id

    @property
    def wh(self) -> Tuple[int, int]:
        assert self.content_type == "image"
        if isinstance(self.content, np.ndarray):
            h, w = self.content.shape[:2]
        elif isinstance(self.content, Image.Image):
            w, h = self.content.size
        else:
            raise ValueError(f"Invalid image type, got {type(self.content)}")
        return w, h

    def resize(self, width: int, height: int) -> "DataFile":
        assert self.content_type == "image"
        self.content = resize_image(self.content, width=width, height=height)
        return self

    def crop_content(
        self,
        locations: Union[
            List[Polygon], List[Tuple[int, int, int, int]], List[Tuple[float, float, float, float]]
        ],
        crop_args: Optional["CropArgs"] = None,
    ) -> List[Union[np.ndarray, Image.Image]]:
        from highlighter.datasets.cropping import crop_rect_from_poly

        assert self.content_type == "image"

        if isinstance(locations[0], tuple) and isinstance(locations[0][0], float):
            ori_w, ori_h = self.wh
            polys = [
                Polygon(
                    [
                        (int(x0 * ori_w), int(y0 * ori_h)),
                        (int(x1 * ori_w), int(y0 * ori_h)),
                        (int(x1 * ori_w), int(y1 * ori_h)),
                        (int(x0 * ori_w), int(y1 * ori_h)),
                        (int(x0 * ori_w), int(y0 * ori_h)),
                    ]
                )
                for x0, y0, x1, y1 in locations
            ]
        elif isinstance(locations[0], tuple) and isinstance(locations[0][0], int):
            ori_w, ori_h = self.wh
            polys = [
                Polygon(
                    [
                        (x0, y0),
                        (x1, y0),
                        (x1, y1),
                        (x0, y1),
                        (x0, y0),
                    ]
                )
                for x0, y0, x1, y1 in locations
            ]
        elif isinstance(locations[0], Polygon):
            polys = locations
        else:
            raise ValueError(f"Invalid crop locations, got, {locations}.")
        return [crop_rect_from_poly(self.content, p, crop_args) for p in polys]

    def draw_annotations(self, annotations: "Annotations"):
        overlay = draw_annotations_on_image(self.content, annotations)
        return overlay

    class ContentTypeEnum(Enum):
        UNKNOWN = "unknown"
        IMAGE = "image"
        VIDEO = "video"
        TEXT = "text"
        JSON = "json"
        AUDIO = "audio"
        WEB_PAGE = "web_page"
        KML = "kml"



"""
thread_watch.py – drop-in helper to log every Thread that is created.

• Counts live threads in real time
• Records the file/line that called Thread(...)
• Survives 3rd-party libs because it replaces threading.Thread globally
"""

import atexit
import logging
import signal
import sys

# ─── thread_watch.py ─────────────────────────────────────────────────────
import threading
import traceback
import weakref

# Public, process-wide registry ───────────────────────────────────────────
active_threads: "weakref.WeakSet[threading.Thread]" = weakref.WeakSet()
_callsites: "weakref.WeakKeyDictionary[threading.Thread, tuple[str, list[str]]]" = weakref.WeakKeyDictionary()
_lock = threading.Lock()

# Original methods we patch around
_orig_init = threading.Thread.__init__
_orig_start = threading.Thread.start


logger = logging.getLogger(__name__)


def _init(self, *a, **kw):
    _orig_init(self, *a, **kw)
    stack = traceback.format_stack(limit=6)[:-2]
    _callsites[self] = (self.name, stack)


def _start(self, *a, **kw):
    with _lock:
        active_threads.add(self)  # add thread to registry
    logger.info("[+THREAD] %s  total=%d", self.name, len(active_threads))
    return _orig_start(self, *a, **kw)


def _run_wrapper(orig_run):
    def wrapped(self, *a, **kw):
        try:
            return orig_run(self, *a, **kw)
        finally:
            with _lock:
                active_threads.discard(self)
            logger.info("[-THREAD] %s  total=%d", self.name, len(active_threads))

    return wrapped


threading_patched = False


def patch_threading():
    """Monkey patch threading to register threads and dump on exit"""
    global threading_patched
    if threading_patched:
        return
    threading_patched = True
    threading.Thread.__init__ = _init
    threading.Thread.start = _start
    threading.Thread.run = _run_wrapper(threading.Thread.run)


# ─────────── public utility helpers ──────────────────────────────────────
def live_threads() -> list[threading.Thread]:
    """Return a *snapshot* list of threads still alive (excludes current)."""
    with _lock:
        return [t for t in active_threads if t.is_alive() and t is not threading.current_thread()]


def join_all(timeout: float | None = None) -> None:
    """Block until every registered thread finishes."""
    for t in live_threads():
        t.join(timeout)


def dump_stacks(level=logging.WARNING) -> None:
    for t in live_threads():
        name, stack = _callsites.get(t, ("<?>", []))
        logger.log(level, "THREAD STILL LIVE: %s (ident=%s)\n%s", name, t.ident, "".join(stack))


# Keep the atexit dump as a last resort
@atexit.register
def _atexit():
    if live_threads():
        logger.warning("Process exiting with %d worker thread(s) alive!", len(live_threads()))
        dump_stacks()



from typing import Optional
from uuid import UUID

from ...core import GQLBaseModel

__all__ = ["User"]


class User(GQLBaseModel):
    id: int
    display_name: Optional[str] = None
    email: str
    uuid: Optional[UUID] = None



import json
from typing import List, Optional

import click

from ..client import HLJSONEncoder
from ..client.base_models.base_models import CaseType
from ..core import GQLBaseModel


class CreateCasePayload(GQLBaseModel):
    errors: List[str]
    case: Optional[CaseType] = None


@click.group("case")
@click.pass_context
def case_group(ctx):
    pass


@case_group.command("create")
@click.option(
    "-w",
    "--workflow-order-id",
    type=str,
    required=True,
    help="Workflow order ID",
)
@click.option(
    "-n",
    "--name",
    type=str,
    required=False,
    help="Case name",
)
@click.option(
    "-e",
    "--entity-id",
    type=str,
    required=False,
    help="Entity ID",
)
@click.option(
    "-f",
    "--initial-data-file-ids",
    type=str,
    required=False,
    help="Initial data file IDs (comma-separated)",
    multiple=True,
)
@click.pass_context
def create(ctx, workflow_order_id, name, entity_id, initial_data_file_ids):
    """Create a new case"""
    client = ctx.obj["client"]

    # Flatten, split, and filter comma-separated IDs
    file_ids = [
        fid.strip() for id_group in initial_data_file_ids for fid in id_group.split(",") if fid.strip()
    ]

    result = client.createCase(
        return_type=CreateCasePayload,
        workflowOrderId=workflow_order_id,
        name=name,
        entityId=entity_id,
        initialDataFileIds=file_ids if file_ids else None,
    ).gql_dict()

    click.echo(json.dumps(result, cls=HLJSONEncoder))

    # Exit with non-zero code if there are errors
    if result.get("errors"):
        raise click.ClickException("Case creation failed")



# Used when there is no connected transport layer, e.g no MQTT !

from typing import Any

from aiko_services.main.message import *

__all__ = ["Castaway"]

class Castaway(Message):
    def __init__(
        self: Any,
        message_handler: Any = None,
        topics_subscribe: Any = None,
        topic_lwt: str = None,
        payload_lwt:str = None,
        retain_lwt: bool = False
        ) -> None:

        pass

    def publish(
        self: Any,
        topic: str,
        payload: Any,
        retain: bool = False,
        wait: bool = False
        ) -> None:

        pass

    def set_last_will_and_testament(
        self: Any,
        topic_lwt: str = None,
        payload_lwt: str = "(absent)",
        retain_lwt: bool = False
        ) -> None:

        pass

    def subscribe(self: Any, topics: Any) -> None:
        pass

    def unsubscribe(self: Any, topics: Any, remove=True) -> None:
        pass

# Wilson !



#!/usr/bin/env python3
"""
CLI entrypoint for RTSP server.

Usage:
  rtsp-server [video_file] [--port 8554] [--host 0.0.0.0] [--advertise-host 100.x.y.z] [--tcp-only] [--no-loop]
"""

import argparse
import sys
from pathlib import Path

# Try relative import first (for installed package), fallback to direct import (for standalone use)
try:
    from .rtsp_server import RTSPServer
except ImportError:
    import importlib.util

    spec = importlib.util.spec_from_file_location("rtsp_server", Path(__file__).parent / "rtsp_server.py")
    rtsp_server = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(rtsp_server)
    RTSPServer = rtsp_server.RTSPServer

# Default test video file - resolve to absolute path
DEFAULT_VIDEO_FILE = (
    Path(__file__).parent / "../../../../../web/spec/files/videos/sample_video_360x240x15FPS_5s.mp4"
).resolve()
TEST_VIDEOS_DIR = (Path(__file__).parent / "../../../../../web/spec/files/videos").resolve()


def main():
    parser = argparse.ArgumentParser(description="GStreamer RTSP Server for testing")
    parser.add_argument(
        "video_file", nargs="?", default=str(DEFAULT_VIDEO_FILE), help="Path to MP4 video file"
    )
    parser.add_argument("--port", "-p", type=int, default=8554, help="RTSP server port (default: 8554)")
    parser.add_argument(
        "--host",
        default="0.0.0.0",  # nosec B104
        help="Bind address (use 0.0.0.0 for all interfaces; default: 0.0.0.0)",
    )
    parser.add_argument(
        "--advertise-host",
        default=None,
        help="Host/IP to show in the printed RTSP URL (e.g., your Tailscale IP like 100.x.y.z)",
    )
    parser.add_argument(
        "--tcp-only",
        action="store_true",
        help="Force RTSP over TCP (interleaved) to keep everything on one TCP connection (recommended over Tailscale/NAT).",
    )
    parser.add_argument(
        "--no-loop",
        action="store_true",
        help="Disable looping - stream will end when video file finishes (default: looping enabled)",
    )

    args = parser.parse_args()

    video_path = Path(args.video_file)
    if not video_path.exists():
        print(f"Error: Video file not found: {video_path}")
        print("\nAvailable test videos:")
        if TEST_VIDEOS_DIR.exists():
            for mp4_file in TEST_VIDEOS_DIR.glob("*.mp4"):
                print(f"  {mp4_file}")
        sys.exit(1)

    # Create and run RTSP server
    server = RTSPServer(
        video_file=video_path,
        port=args.port,
        bind_address=args.host,
        advertise_host=args.advertise_host,
        tcp_only=args.tcp_only,
        loop=not args.no_loop,
    )
    server.run()


if __name__ == "__main__":
    main()



import json
from pathlib import Path
from typing import Dict, List, Optional

import click
import yaml

from ..client import (
    HLJSONEncoder,
    TrainingRunArtefactType,
    TrainingRunArtefactTypeEnum,
    TrainingRunType,
    create_capability_for_artefact,
    get_latest_training_run_artefact,
    upload_file_to_s3,
)
from ..core import DEPRECATED_CAPABILITY_IMPLEMENTATION_FILE, GQLBaseModel
from .common import _to_pathlib


def load_config(cfg_path: Optional[Path]) -> Dict:
    if cfg_path is None:
        return {}

    loaders = {
        ".yaml": yaml.safe_load,
        ".json": json.load,
    }

    with cfg_path.open("r") as f:
        cfg = loaders[cfg_path.suffix](f)

    if cfg.get("parameters", {}).get("cropper", {}).get("scale", 1.0) != 1.0:
        raise ValueError(
            "Got'cha!!!, Dont push an artefact with " "cropping scale other than 1.0 or remove from dict"
        )

    return cfg


@click.group("training-run")
@click.pass_context
def training_run_group(ctx):
    pass


@training_run_group.group("artefact")
@click.pass_context
def artefact_group(ctx):
    pass


@artefact_group.command("read")
@click.option(
    "-i",
    "--id",
    type=str,
    required=True,
)
@click.option(
    "-s",
    "--save-path",
    type=click.Path(file_okay=True, writable=True),
    required=True,
    callback=_to_pathlib,
)
@click.option(
    "-a",
    "--artefact-type",
    type=click.Choice(
        list(TrainingRunArtefactTypeEnum.__members__.keys()) + [DEPRECATED_CAPABILITY_IMPLEMENTATION_FILE]
    ),
    required=True,
)
@click.pass_context
def read_artefact(ctx, id, save_path, artefact_type):
    client = ctx.obj["client"]

    artefact = get_latest_training_run_artefact(
        client,
        id,
        download_file_url=True,
        file_url_save_path=save_path,
        filter_by_artefact_type=artefact_type,
    )
    assert artefact is not None
    artefact_yaml_path = save_path.parent / f"{artefact.id}.yaml"
    artefact.dump_yaml(artefact_yaml_path)
    click.echo(json.dumps(artefact.model_dump(), cls=HLJSONEncoder))


@artefact_group.command("create")
@click.option(
    "-i",
    "--id",
    type=str,
    required=True,
    help="training run id",
)
@click.option(
    "-a",
    "--artefact-yaml",
    type=click.Path(file_okay=True, dir_okay=False, exists=True),
    required=True,
    callback=_to_pathlib,
    help=".yaml file base_model.py:TrainingRunArtefactType",
)
@click.pass_context
def create_artefact(ctx, id, artefact_yaml):
    """Create an artefact for a training run and upload artefact-yaml.file_url to s3.

    \b
    Fields in artefact.yaml:
    - file_url: should contain the absolute path to the checkpoint file you want to upload.
    - type: [REQUIRED] one of OnnxOpset11, OnnxOpset14, OnnxRuntimeAmd64, OnnxRuntimeArm, TorchScriptV1
    - checkpoint: path to checkpoint in file
    - inference_config: [REQUIRED] inference configuration in json format
    - training_config: [REQUIRED] training configuration in json format

    \b
    # artefact.yaml
    file_url: /home/users/rick/checkpoint.onnx14
    type: OnnxOpset14
    inference_config: {}
    training_config: {}
    """
    client = ctx.obj["client"]

    artefact: TrainingRunArtefactType = TrainingRunArtefactType.from_yaml(artefact_yaml)
    artefact.create(client, id)


@training_run_group.command("create")
@click.option(
    "-eid",
    "--evaluation-id",
    type=str,
    required=True,
)
@click.option(
    "-eid",
    "--experiment-id",
    type=str,
    required=True,
)
@click.option(
    "-mid",
    "--capability-id",
    type=str,
    required=True,
)
@click.option(
    "-pid",
    "--workflow-id",
    type=str,
    required=True,
)
@click.option(
    "-n",
    "--name",
    type=str,
    required=True,
)
@click.option(
    "--source-code-url",
    type=str,
    required=False,
)
@click.option(
    "--source-code-commit-hash",
    type=str,
    required=False,
)
@click.option(
    "--training-log-archive",
    type=click.Path(dir_okay=False),
    callback=_to_pathlib,
    required=False,
)
@click.pass_context
def create_training_run(
    ctx,
    evaluation_id,
    experiment_id,
    capability_id,
    workflow_id,
    name,
    source_code_url,
    source_code_commit_hash,
    training_log_archive,
):
    client = ctx.obj["client"]

    class CreateTrainingRunPayload(GQLBaseModel):
        errors: List[str]
        training_run: TrainingRunType

    result = client.createTrainingRun(
        return_type=CreateTrainingRunPayload,
        researchPlanId=evaluation_id,
        experimentId=experiment_id,
        modelId=capability_id,
        projectId=workflow_id,
        name=name,
        sourceCodeUrl=source_code_url,
        sourceCodeCommitHash=source_code_commit_hash,
    )
    print(result.training_run.id)



import logging
import os
import sys
import warnings
from typing import Optional

import click

from highlighter.core.config import (
    HighlighterRuntimeConfig,
    HighlighterRuntimeConfigError,
)

from ..client import HLClient
from ..client.gql_client import (
    CONST_DEFAULT_GRAPHQL_PROFILES_YAML,
    ENV_HL_WEB_GRAPHQL_API_TOKEN,
    ENV_HL_WEB_GRAPHQL_ENDPOINT,
)
from .agent import agent_group
from .assessment import assessment_group
from .case import case_group
from .config import config_group
from .data_files import data_file_group
from .dataset import dataset_group
from .evaluation import evaluation_group
from .experiment import experiment_group
from .logging import configure_root_logger
from .object_class import object_class_group
from .step import step_group
from .task import task_group
from .template import generate_group
from .trainer import train_group
from .training_run import training_run_group


class NoHLClientCredentialsError(Exception):
    def __init__(self):
        message = (
            "\nNo way of determining credentials for HLClient "
            "could be found. \n"
            "\tOption 1: Use the --profile flag in the cli\n"
            "\tOption 2: Use the --api-token and --endpoint-url flags in the cli\n"
            "\tOption 3: export environment variables"
        )
        super().__init__(message)


class NoHLClient:
    """
    Fallback class for when a user does not provide a way
    of deterimining Highlighter credentials. This allows them
    to run help commands and other commands like 'config' that
    do not rely on a working HLClient, but will give a nice
    error message when they do.
    """

    def __getattr__(self, key):
        raise NoHLClientCredentialsError()

    def __repr__(self):
        return "NoHLClient"


def init_cli_context(
    ctx,
    api_token: Optional[str],
    endpoint_url: Optional[str],
    profile: Optional[str],
    profiles_path: Optional[str],
    highlighter_config_path: Optional[str],
):
    if ctx.obj is None:
        ctx.obj = {}

    logger = logging.getLogger(__name__)

    # Init HLClient
    if profile is not None:
        client = HLClient.from_profile(profile=profile, profiles_path=profiles_path)
    elif (endpoint_url is not None) and (api_token is not None):
        client = HLClient.from_credential(
            api_token=api_token,
            endpoint_url=endpoint_url,
        )
    elif ENV_HL_WEB_GRAPHQL_ENDPOINT in os.environ:
        client = HLClient.get_client()
    else:
        client = NoHLClient()

    ctx.obj.update({"client": client})
    logger.debug(f"HLClient: {client}")

    # Load HighlighterRuntimeConfig
    try:
        hl_cfg = HighlighterRuntimeConfig.load(config_path=highlighter_config_path)
        ctx.obj.update({"hl_cfg": hl_cfg})
        logger.debug(f"Loaded HighlighterRuntimeConfig from: {hl_cfg.config_path}")
    except (HighlighterRuntimeConfigError, ValueError) as e:
        logger.error(f"Failed to load configuration: {e}")
        sys.exit(2)
    except Exception as e:
        logger.error(f"Unexpected error during loading configuration: {e}")
        sys.exit(1)


@click.group("highlighter")
@click.option("--api-token", type=str)
@click.option("--endpoint-url", type=str)
@click.option("--profile", type=str, default=None)
@click.option("--profiles-path", type=str, default=CONST_DEFAULT_GRAPHQL_PROFILES_YAML)
@click.option("--highlighter-config-path", type=str, default=None)
@click.pass_context
def highlighter_group(ctx, api_token, endpoint_url, profile, profiles_path, highlighter_config_path):
    init_cli_context(ctx, api_token, endpoint_url, profile, profiles_path, highlighter_config_path)
    configure_root_logger(
        _log_path=ctx.obj["hl_cfg"].log_path,
        _log_level=ctx.obj["hl_cfg"].log_level,
        _log_rotation_max_kilobytes=ctx.obj["hl_cfg"].log_rotation_max_kilobytes,
        _log_rotation_backup_count=ctx.obj["hl_cfg"].log_rotation_backup_count,
    )
    warnings.simplefilter("default")  # Suppress duplicated warnings


@highlighter_group.command("write")
@click.argument("outfile", type=str)
@click.pass_context
def write(ctx, outfile):
    """Write credentials to a file.

    Each credential is appended to the given file.

    eg: highlighter-v2 --profile abc .envrc

    Results in the following lines being appended
    the .envrc

    export HL_WEB_GRAPHQL_ENDPOINT=...
    export HL_WEB_GRAPHQL_API_TOKEN=...
    """
    client = ctx.obj["client"]
    client.append_credentials_to_env_file(outfile)


@highlighter_group.command("export")
@click.pass_context
def export(ctx):
    """Export a profile's credentials to env

    Wrap this command in `back ticks` to export a profile's credentials to
    your environment

    eg: `highlighter-v2 --profile abc export`

    Results in the following credentials being
    added to your environment variables

    HL_WEB_GRAPHQL_ENDPOINT=...
    HL_WEB_GRAPHQL_API_TOKEN=...
    """
    client = ctx.obj["client"]

    click.echo("Wrap the command in `back ticks` to execute the exports", err=True)
    click.echo(
        f"export {ENV_HL_WEB_GRAPHQL_ENDPOINT}={client.endpoint_url} {ENV_HL_WEB_GRAPHQL_API_TOKEN}={client.api_token}"
    )


highlighter_group.add_command(agent_group)
highlighter_group.add_command(assessment_group)
highlighter_group.add_command(case_group)
highlighter_group.add_command(config_group)
highlighter_group.add_command(data_file_group)
highlighter_group.add_command(data_file_group, name="file")  # Alias for data-file
highlighter_group.add_command(dataset_group)
highlighter_group.add_command(experiment_group)
highlighter_group.add_command(object_class_group)
highlighter_group.add_command(step_group)
highlighter_group.add_command(task_group)
highlighter_group.add_command(training_run_group)
highlighter_group.add_command(evaluation_group)
highlighter_group.add_command(generate_group)
highlighter_group.add_command(train_group)


if __name__ == "__main__":
    highlighter_group()



from sqlalchemy.engine import Engine
from sqlmodel import Session, SQLModel, create_engine

from highlighter.core.config import HighlighterRuntimeConfig

__all__ = [
    "Database",
]


class Database:
    def __init__(self):
        hl_cfg = HighlighterRuntimeConfig.load()
        self.highlighter_path_to_database_file = str(hl_cfg.agent.db_file())
        self._engine = create_engine(
            f"sqlite:///{self.highlighter_path_to_database_file}", connect_args={"check_same_thread": False}
        )
        SQLModel.metadata.create_all(self._engine)

    @property
    def engine(self) -> Engine:
        return self._engine

    def get_session(self):
        return Session(self.engine)

    def close(self):
        self._engine.dispose()  # Call this when shutting down your app



from typing import Dict, List
from uuid import UUID

from highlighter.core.enums import ContentTypeEnum
from highlighter.core.gql_base_model import GQLBaseModel

from .entity import Entity


class Entities(GQLBaseModel):
    """Entity container

    Enables erganomic management of a set of entities.
    Entities can be looked-up by ID:
        `entity = entities[entity_id]`
    Entities can be added:
        `entities[entity_id] = entity`
        `entities.add(entity)`
        `entities.update(other_entities)`
    Entities can be queried (not yet implemented):
        `specific_entities = entities.where(object_class=object_class_id)`
        `specific_entities = entities.where(has_attribute=attribute_id)`
        `specific_entities = entities.where(has_attribute_value=enum_id)`
    """

    _entities: Dict[UUID, Entity] = {}

    def add(self, entity: Entity):
        self._entities[entity.id] = entity

    def __getitem__(self, key: UUID | int):
        if isinstance(key, int):
            return list(self._entities.values())[key]
        return self._entities[key]

    def __delitem__(self, key: UUID):
        return self._entities.__delitem__(key)

    def remove(self, entity: Entity):
        del self[entity.id]

    def __len__(self) -> int:
        return len(self._entities)

    def __iter__(self):
        return iter(list(self._entities.values()))

    def get(self, key: UUID, default: Entity | None = None):
        return self._entities.get(key, default)

    def __setitem__(self, entity_id: UUID, entity: Entity):
        self._entities[entity_id] = entity

    def update(self, *args, **kwargs):
        # Handles both dicts and iterable of pairs
        if args:
            other = args[0]
            if hasattr(other, "items"):
                for k, v in other.items():
                    self[k] = v  # goes through __setitem__
            else:
                for value in other:
                    if isinstance(value, Entity):
                        self.add(value)
                    else:
                        k, v = value
                        self[k] = v
        for k, v in kwargs.items():
            self[k] = v

    def __ior__(self, other):
        self.update(other)
        return self

    def __or__(self, other):
        new = type(self)(self, _entities=self._entities.copy())
        new.update(other)
        return new

    def __repr__(self):
        return self._entities.__repr__()

    def to_json_serializable_dict(self):
        return {str(id): entity.to_json() for id, entity in self._entities.items()}

    def to_data_sample(self) -> "DataSample":
        from highlighter.core.data_models.data_sample import DataSample

        if len(self._entities) == 0:
            return DataSample(content=self, content_type=ContentTypeEnum.ENTITIES)
        some_annotations = list(self._entities.values())[0].annotations
        if len(some_annotations) == 0:
            raise ValueError("Cannot convert Entities to DataSample if there are no annotations")
        annotation = some_annotations[0]
        if annotation.data_file_id is None:
            raise ValueError("Cannot convert Entities to DataSample if annotation.data_file_id is None")
        return DataSample(
            content=self,
            content_type=ContentTypeEnum.ENTITIES,
            recorded_at=annotation.occurred_at,
            stream_frame_index=annotation.datum_source.frame_id,
            media_frame_index=annotation.datum_source.frame_id,  # FIXME
        )
